/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "debugl.h"
#include "tools.h"
#include "string.h"

#ifdef HAVE_DEBUG_FEATURES

const debug_level_type cfg_debug_levels[] = 
{
    { ~0UL, "all", NULL },
    { DLV1, "html", gettext_nop("HTML parsers") },
    { DLV2, "protos", gettext_nop("Server responses") },
    { DLV3, "protoc", gettext_nop("Client requests") },
    { DLV4, "procs", gettext_nop("Procedure calling") },
    { DLV5, "locks", gettext_nop("File locking") },
    { DLV6, "net", gettext_nop("Networking code") },
    { DLV7, "misc", gettext_nop("Miscelanous") },
    { DLV8, "user", gettext_nop("Extended user infos") },
    { DLV9, "mtlock", gettext_nop("Multithreading - locking") },
    { DLV10, "mtthr", gettext_nop("Multithreading - threads") },
    { DLV11, "protod", gettext_nop("POST request data") },
    { DLV12, "limits", gettext_nop("Limiting conditions") },
    { DLV13, "ssl", gettext_nop("SSL informations") },
//#ifdef INCLUDE_CHUNKY_DoS_FEATURES
    { DLV14, "hammer", gettext_nop("HAMMER action info") },
//#endif
    { DLV15, "trace", gettext_nop("TRACE data for a developer") },
    { DLV16, "dev", gettext_nop("developer info - should be moved to other levels ASAP") },
    { DLV17, "bufio", gettext_nop("developer info - bufio handling") },
    { DLV18, "cookie", gettext_nop("cookie handling") },
    { DLV19, "htmlform", gettext_nop("html FORM processing") },
    { DLV20, "rules", gettext_nop("fnrules and JavaScript processing") },
    { DLV21, "js", gettext_nop("JavaScript pattern processing") },
    { 0, NULL, NULL }
};

int debug_level_parse(const char *str, unsigned long *dst)
{
    unsigned int dl = 0;
    const char *p = str;
    int l, i;
    bool_t found;

    ASSERT(dst != NULL);
    *dst = 0;
    while (*p)
    {
        p += strspn(p, ",");
        l = strcspn(p, ",");
        found = FALSE;
        for (i = 0; cfg_debug_levels[i].id; i++)
        {
            if (!strncasecmp(p, cfg_debug_levels[i].option, l) && (l == strlen(cfg_debug_levels[i].option)))
            {
                found = TRUE;
                dl |= cfg_debug_levels[i].id;
                break;
            }
            else if (*p == '!' && (!strncasecmp(p + 1, cfg_debug_levels[i].option, l - 1) && (l - 1 == strlen(cfg_debug_levels[i].option))))
            {
                found = TRUE;
                dl &= ~cfg_debug_levels[i].id;
                break;
            }
        }
        if (!found)
        {
            char *ep;

            i = strtol(p, &ep, 0);

            if ((ep - p) != l)
            {
                xprintf(0, gettext("Bad debug level selection: %s. Cannot decode '%s' remainder\n"), str, p);
                return -1;
            }
            else
            {
                dl |= i;
            }
        }
        p += l;
    }

    *dst = dl;
    return 0;
}

char *debug_level_construct(unsigned long level)
{
    bool_t frst = TRUE;
    int i;
    char *strbuf = NULL;
    unsigned long orig_level = level;

    if (!level)
    {
    return tl_strdup("0");
    }

    for (i = 0; cfg_debug_levels[i].id; i++)
    {
        if ((cfg_debug_levels[i].id & level) != cfg_debug_levels[i].id)
            continue;

        if (!frst)
            strbuf = tl_str_concat(DBGvars(strbuf), ",", NULL);
        strbuf = tl_str_concat(DBGvars(strbuf), cfg_debug_levels[i].option, NULL);
        frst = FALSE;
        level &= ~cfg_debug_levels[i].id;
    }
    if (level)
    {
    char *sbufbuf = strbuf;

        /* if any bits are left, try to encode the levels as 'all,!x,!y,!z,!... first: */
    level = orig_level ^ ~0;
        ASSERT(level != 0);

        strbuf = tl_strdup("all");
        for (i = 0; cfg_debug_levels[i].id; i++)
        {
            if ((cfg_debug_levels[i].id & level) != cfg_debug_levels[i].id)
                continue;
            strbuf = tl_str_concat(DBGvars(strbuf), ",!", cfg_debug_levels[i].option, NULL);
            level &= ~cfg_debug_levels[i].id;
        }
        if (level)
        {
            char buf[20];

      _free(strbuf);
      strbuf = sbufbuf;
            sprintf(buf, "%lu", level);
      if(strbuf && *strbuf)
        strbuf = tl_str_concat(DBGvars(strbuf), ",", NULL);
      strbuf = tl_str_concat(DBGvars(strbuf), buf, NULL);
    }
    else
    {
      _free(sbufbuf);
        }
    }
    return strbuf;
}





void DEBUG_HTML(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV1), "[HTML]", str, &args);
    va_end(args);
#endif
}

void DEBUG_PROTOS(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV2), "[PROTOS]  ", str, &args);
    va_end(args);
#endif
}

void DEBUG_PROTOC(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV3), "[PROTOC]  ", str, &args);
    va_end(args);
#endif
}

void DEBUG_PROCS__(const char *source, int lineno, const char *func, const char *str)
{
#ifdef HAVE_DEBUG_FEATURES
    xdebug(DBGnull(DLV4), "[PROCS]   ", "(%s:%d) calling - %s() %s\n", source, lineno, func, (str ? str : ""));
#endif
}

void DEBUG_PROCE__(const char *source, int lineno, const char *func, const char *str)
{
#ifdef HAVE_DEBUG_FEATURES
    xdebug(DBGnull(DLV4), "[PROCE]   ", "(%s:%d) exiting - %s() %s\n", source, lineno, func, (str ? str : ""));
#endif
}

void DEBUG_LOCKS(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV5), "[LOCKS]   ", str, &args);
    va_end(args);
#endif
}

void DEBUG_NET(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV6), "[NET]     ", str, &args);
    va_end(args);
#endif
}

void DEBUG_MISC(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV7), "[MISC]    ", str, &args);
    va_end(args);
#endif
}

void DEBUG_USER(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV8), "[USER]    ", str, &args);
    va_end(args);
#endif
}

void DEBUG_MTLOCK(DBGdecl(const char *str), ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGpass(DLV9), "[MTLOCK]  ", str, &args);
    va_end(args);
#endif
}

void DEBUG_MTTHR(DBGdecl(const char *str), ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGpass(DLV10), "[MTTHR]   ", str, &args);
    va_end(args);
#endif
}

void DEBUG_PROTOD(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV11), "[PROTOD]  ", str, &args);
    va_end(args);
#endif
}

void DEBUG_LIMITS(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV12), "[LIMITS]  ", str, &args);
    va_end(args);
#endif
}
void DEBUG_SSL(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV13), "[SSL]     ", str, &args);
    va_end(args);
#endif
}

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
void DEBUG_HAMMER(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV14), "[HAMMER]  ", str, &args);
    va_end(args);
#endif
}
#endif

void DEBUG_TRACE(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV15), "[TRACE]   ", str, &args);
    va_end(args);
#endif
}

void DEBUG_DEVEL(DBGdecl(const char *str), ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGpass(DLV16), "[DEV]     ", str, &args);
    va_end(args);
#endif
}

void DEBUG_BUFIO(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV17), "[BUFIO]   ", str, &args);
    va_end(args);
#endif
}

void DEBUG_COOKIE(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV18), "[COOKIE]  ", str, &args);
    va_end(args);
#endif
}

void DEBUG_HTMLFORM(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV19), "[HTMLFORM]", str, &args);
    va_end(args);
#endif
}

void DEBUG_RULES(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV20), "[RULES]   ", str, &args);
    va_end(args);
#endif
}

void DEBUG_JS(const char *str, ...)
{
#ifdef HAVE_DEBUG_FEATURES
    va_list args;
    va_start(args, str);
    xvdebug(DBGnull(DLV21), "[JS]      ", str, &args);
    va_end(args);
#endif
}

#endif /* HAVE_DEBUG_FEATURES */

