/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "gui.h"
#include "gui_api.h"

#ifdef GTK_FACE

#if !(defined(__CYGWIN__) || defined(WIN32) || defined(__WIN32))
#include <X11/Xlib.h>
#include <gdk/gdkx.h>
#endif

#include "tools.h"

#include "icons.h"
#include "gui_api.h"
#include "gkeys.h"
#include "gaccel.h"



void gui_beep(void)
{
    DEBUG_PROCS(NULL);

    gdk_beep();

    DEBUG_PROCE(NULL);
}

void gui_start_download(void)
{
    DEBUG_PROCS(NULL);

    if (!MT_IS_MAIN_THREAD())
    {
        DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
        gdk_threads_enter();
    }

    gtk_label_set(GTK_LABEL(gui_cfg.status_msg), "");
    gtk_widget_set_sensitive(gui_cfg.bt_restart, FALSE);
    gtk_widget_set_sensitive(gui_cfg.mea_restart, FALSE);
    gtk_widget_set_sensitive(gui_cfg.mtb_restart, FALSE);
    gtk_widget_set_sensitive(gui_cfg.bt_continue, FALSE);
    gtk_widget_set_sensitive(gui_cfg.mea_continue, FALSE);
    gtk_widget_set_sensitive(gui_cfg.mtb_continue, FALSE);
    gtk_widget_set_sensitive(gui_cfg.bt_stop, TRUE);
    gtk_widget_set_sensitive(gui_cfg.mea_stop, TRUE);
    gtk_widget_set_sensitive(gui_cfg.mtb_stop, TRUE);
    gtk_widget_set_sensitive(gui_cfg.bt_break, TRUE);
    gtk_widget_set_sensitive(gui_cfg.mea_break, TRUE);
    gtk_widget_set_sensitive(gui_cfg.mtb_break, TRUE);

#ifdef WITH_TREE
    gtk_label_set(GTK_LABEL(gui_cfg.tree_help), "");
#endif

    if (!MT_IS_MAIN_THREAD())
    {
        DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
        gdk_flush();            /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
        DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
        gdk_threads_leave();
    }

    gui_loop_serve();

    DEBUG_PROCE(NULL);
}

void gui_finish_download(void)
{
    DEBUG_PROCS(NULL);

    cfg.processing = FALSE;

    if (!MT_IS_MAIN_THREAD())
    {
        DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
        gdk_threads_enter();
    }

    gdk_beep();

    gtk_widget_set_sensitive(gui_cfg.bt_restart, TRUE);
    gtk_widget_set_sensitive(gui_cfg.mea_restart, TRUE);
    gtk_widget_set_sensitive(gui_cfg.mtb_restart, TRUE);
    if (cfg.urlstack)
    {
        gtk_widget_set_sensitive(gui_cfg.bt_continue, TRUE);
        gtk_widget_set_sensitive(gui_cfg.mea_continue, TRUE);
        gtk_widget_set_sensitive(gui_cfg.mtb_continue, TRUE);
    }
    gtk_widget_set_sensitive(gui_cfg.bt_stop, FALSE);
    gtk_widget_set_sensitive(gui_cfg.mea_stop, FALSE);
    gtk_widget_set_sensitive(gui_cfg.mtb_stop, FALSE);
    gtk_widget_set_sensitive(gui_cfg.bt_break, FALSE);
    gtk_widget_set_sensitive(gui_cfg.mea_break, FALSE);
    gtk_widget_set_sensitive(gui_cfg.mtb_break, FALSE);

    if (!MT_IS_MAIN_THREAD())
    {
        DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
        gdk_flush();            /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
        DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
        gdk_threads_leave();
    }

    gui_loop_serve();

    cfg.rbreak = FALSE;
    cfg.stop = FALSE;

    DEBUG_PROCE(NULL);
}

void gui_finish_document(doc *docp)
{
    DEBUG_PROCS(NULL);

#ifdef WITH_TREE
    if (cfg.xi_face)
    {
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        if (GTK_TOGGLE_BUTTON(gui_cfg.watch_download)->active)
        {
            LOCK_GTKTREE;
            gtk_ctree_select(GTK_CTREE(gui_cfg.tree_widget), docp->doc_url->tree_nfo[0]);
            gtk_ctree_node_moveto(GTK_CTREE(gui_cfg.tree_widget), docp->doc_url->tree_nfo[0], 0, 0.0, 0.0);
            UNLOCK_GTKTREE;
        }

        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }

        gui_loop_serve();
    }
#endif

    DEBUG_PROCE(NULL);
}


void tl_enter_fn(void)
{
    LOCK_GTK;
}

void tl_leave_fn(void)
{
    UNLOCK_GTK;
}

void gui_start(int *argc, char **argv)
{
    char *p;
    char gtkrc[PATH_MAX];

    DEBUG_PROCS(NULL);

#if defined(__CYGWIN__) || defined(WIN32) || defined(__WIN32)
    snprintf(gtkrc, sizeof(gtkrc), "%s%s%s-gtkrc", cfg.install_path, (tl_is_dirname(cfg.install_path) ? "" : "/"), PACKAGE);
    gtkrc[sizeof(gtkrc) - 1] = 0;
    gtk_rc_add_default_file(gtkrc);
#endif
    /* Parse the file ~/.pavuk-gtkrc. This makes Pavuk themeable :) */
    if ((p = getenv("HOME")))
    {
    snprintf(gtkrc, sizeof(gtkrc), "%s%s%s%s-gtkrc", p, 
		(tl_is_dirname(p) ? "" : "/"), PAVUK_CFGFNAME_PREFIX, PACKAGE);
        gtkrc[sizeof(gtkrc) - 1] = 0;
        gtk_rc_add_default_file(gtkrc);
    }

    /* see
     * http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html
     *
     * g_thread_init() has already been called in mt_init() before we get here...
     */
#if defined(HAVE_MT)
    gdk_threads_set_lock_functions(tl_enter_fn, tl_leave_fn);
    gdk_threads_init();
#endif
    gtk_init(argc, &argv);
    gkey_load();

    gui_cfg._go_bg = FALSE;
    gui_cfg.toplevel = NULL;
    gui_cfg.about_shell = NULL;
    gui_cfg.scn_load_shell = NULL;
    gui_cfg.scn_save_shell = NULL;
    gui_cfg.cfg_limits = NULL;
    gui_cfg.config_shell = NULL;
    gui_cfg.cfg_sch = NULL;
#ifdef WITH_TREE
    gui_cfg.tree_shell = NULL;
#endif

    DEBUG_PROCE(NULL);
}

void gui_set_doccounter(void)
{
    char pom[256];

    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        sprintf(pom, gettext("Processed: %5ld"), cfg.process_cnt);
        gtk_label_set(GTK_LABEL(gui_cfg.status_done), pom);
        sprintf(pom, gettext("Queued: %5ld"), cfg.total_cnt);
        gtk_label_set(GTK_LABEL(gui_cfg.status_queue), pom);
        sprintf(pom, gettext("Failed: %4ld"), cfg.fail_cnt);
        gtk_label_set(GTK_LABEL(gui_cfg.status_fail), pom);
        sprintf(pom, gettext("Rejected: %5ld"), cfg.reject_cnt);
        gtk_label_set(GTK_LABEL(gui_cfg.status_rej), pom);

        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }

        gui_loop_serve();
    }

    DEBUG_PROCE(NULL);
}

void gui_set_progress(char *sz, char *rate, char *etime, char *rtime)
{
    DEBUG_PROCS(NULL);

#ifdef HAVE_MT
    {
        int thrnr;

        thrnr = cast_voidptr2int(pthread_getspecific(cfg.thrnr_key));

        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 3, sz);
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 4, rate);
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 5, etime);
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 6, rtime);

        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }
    }
#else
    {
        char pom[40];

        snprintf(pom, sizeof(pom), "S: %s", sz);
        pom[sizeof(pom) - 1] = 0;
        gtk_label_set(GTK_LABEL(gui_cfg.status_size), pom);
        snprintf(pom, sizeof(pom), "R: %s", rate);
        pom[sizeof(pom) - 1] = 0;
        gtk_label_set(GTK_LABEL(gui_cfg.status_rate), pom);
        snprintf(pom, sizeof(pom), "ET: %s", etime);
        pom[sizeof(pom) - 1] = 0;
        gtk_label_set(GTK_LABEL(gui_cfg.status_et), pom);
        snprintf(pom, sizeof(pom), "RT: %s", rtime);
        pom[sizeof(pom) - 1] = 0;
        gtk_label_set(GTK_LABEL(gui_cfg.status_rt), pom);
    }
#endif

#if 0
    gui_loop_serve();
#endif

    DEBUG_PROCE(NULL);
}

void gui_clear_status(void)
{
    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
#ifdef HAVE_MT
        int thrnr;

        thrnr = cast_voidptr2int(pthread_getspecific(cfg.thrnr_key));

        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 1, "");
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 2, "");
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 3, "");
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 4, "");
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 5, "");
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 6, "");

        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }
#else
        gtk_entry_set_text(GTK_ENTRY(gui_cfg.minitb_label), "");
        gtk_label_set(GTK_LABEL(gui_cfg.status_size), "");
        gtk_label_set(GTK_LABEL(gui_cfg.status_rate), "");
        gtk_label_set(GTK_LABEL(gui_cfg.status_et), "");
        gtk_label_set(GTK_LABEL(gui_cfg.status_rt), "");
#endif

        gui_loop_serve();
    }

    DEBUG_PROCE(NULL);
}

void gui_set_status(char *text)
{
    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
#ifdef HAVE_MT
        int thrnr;

        thrnr = cast_voidptr2int(pthread_getspecific(cfg.thrnr_key));
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 2, text);

        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }
#else
        gtk_label_set(GTK_LABEL(gui_cfg.status_msg), text);
#endif

        gui_loop_serve();
    }
    else
    {
        DEBUG_USER(" - %s\n", text);
    }

    DEBUG_PROCE(NULL);
}


void gui_set_msg(char *text, int tout)
{
    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_label_set(GTK_LABEL(gui_cfg.status_msg), text);

        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }

        gui_loop_serve();
    }
    else
    {
        DEBUG_USER(" - %s\n", text);
    }

    DEBUG_PROCE(NULL);
}

void gui_set_url(char *str)
{
    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
#ifdef HAVE_MT
        int thrnr;

        thrnr = cast_voidptr2int(pthread_getspecific(cfg.thrnr_key));
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_clist_set_text(GTK_CLIST(gui_cfg.status_list), thrnr, 1, str);

        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }
#else
        gtk_entry_set_text(GTK_ENTRY(gui_cfg.minitb_label), str);
#endif

        gui_loop_serve();
    }

    DEBUG_PROCE(NULL);
}

#ifdef WITH_TREE
void *gui_tree_make_entry(url *urlp)
{
    GtkCTreeNode *retv;
    GtkCTreeNode *parent;
    GdkBitmap *shape;
    GdkPixmap *pixmap;
    gchar *text[1];
    dllist *par;
    url *parurl;

    DEBUG_PROCS(NULL);

    par = dllist_last(urlp->parent_url);

    if (par)
    {
        parurl = (url *)par->data;
    }
    else
    {
        parurl = NULL;
    }

    LOCK_GTKTREE;
    if (!MT_IS_MAIN_THREAD())
    {
        DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
        gdk_threads_enter();
    }
    if (parurl && parurl->tree_nfo && parurl->tree_nfo[0])
    {
        parent = (GtkCTreeNode *)parurl->tree_nfo[0];
    }
    else
    {
        parent = (GtkCTreeNode *)gui_cfg.root;
    }

    g_return_val_if_fail(parent != NULL, NULL);

    if (urlp->status & URL_PROCESSED)
    {
        gchar *text;
        guint8 sp;
        GdkPixmap *p;
        GdkBitmap *m;
        gboolean il, ex;

        gtk_ctree_get_node_info(GTK_CTREE(gui_cfg.tree_widget), urlp->tree_nfo[0], &text, &sp, &p, &m, &pixmap, &shape, &il, &ex);
    }
    else
    {
        pixmap = gui_cfg.icon.notprocessed->pixmap;
        shape = gui_cfg.icon.notprocessed->shape;
    }

    text[0] = url_to_urlstr(urlp, FALSE);

    retv = gtk_ctree_insert_node(GTK_CTREE(gui_cfg.tree_widget), parent, NULL, text, 8, pixmap, shape, pixmap, shape, FALSE, TRUE);

    _free(text[0]);

    gtk_ctree_node_set_row_data(GTK_CTREE(gui_cfg.tree_widget), retv, urlp);

    if (!MT_IS_MAIN_THREAD())
    {
        DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
        gdk_flush();            /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
        DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
        gdk_threads_leave();
    }
    UNLOCK_GTKTREE;

    gui_loop_serve();

    DEBUG_PROCE(NULL);

    return retv;
}
#endif

void gui_create_tree_root_node(void)
{
    DEBUG_PROCS(NULL);

#ifdef WITH_TREE
    if (cfg.xi_face && !gui_cfg.root)
    {
        gchar *text[1];

        text[0] = gettext("Start");

        LOCK_GTKTREE;
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gui_cfg.root =
            (GtkCTreeNode *)gtk_ctree_insert_node(GTK_CTREE(gui_cfg.tree_widget),
                                                  NULL, NULL, text, 8, gui_cfg.icon.notprocessed->pixmap,
                                                  gui_cfg.icon.notprocessed->shape, gui_cfg.icon.notprocessed->pixmap, gui_cfg.icon.notprocessed->shape, FALSE, TRUE);
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_threads_leave();
        }
        UNLOCK_GTKTREE;
    }
#endif /* WITH_TREE */

    DEBUG_PROCE(NULL);
}

void gui_clear_tree(void)
{
    DEBUG_PROCS(NULL);

#ifdef WITH_TREE
    if (cfg.xi_face && gui_cfg.root)
    {
        LOCK_GTKTREE;
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_clist_freeze(GTK_CLIST(gui_cfg.tree_widget));
        gtk_clist_clear(GTK_CLIST(gui_cfg.tree_widget));
        gtk_clist_thaw(GTK_CLIST(gui_cfg.tree_widget));

        gui_cfg.root = NULL;
        if (!MT_IS_MAIN_THREAD())
        {
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }
        UNLOCK_GTKTREE;
    }
#endif /* WITH_TREE */

    DEBUG_PROCE(NULL);
}

void gui_tree_add_start(void)
{
    DEBUG_PROCS(NULL);

#if defined(WITH_TREE) && !defined(HAVE_MT)
    if (cfg.xi_face)
    {
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_clist_freeze(GTK_CLIST(gui_cfg.tree_widget));
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }
    }
#endif

    DEBUG_PROCE(NULL);
}

void gui_tree_add_end(void)
{
    DEBUG_PROCS(NULL);

#if defined(WITH_TREE) && !defined(HAVE_MT)
    if (cfg.xi_face)
    {
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_clist_thaw(GTK_CLIST(gui_cfg.tree_widget));
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }
    }
#endif

    DEBUG_PROCE(NULL);
}

void gui_tree_set_icon_for_doc(doc *docp)
{
    DEBUG_PROCS(NULL);

    icons_set_for_doc(docp);

    DEBUG_PROCE(NULL);
}

#undef gui_loop_serve
void gui_loop_serve(DBGdeclvoid(void))
{
    DEBUG_PROCS(NULL);
    DEBUG_DEVEL(DBGpass("gui_loop_serve() start\n"));

    if (cfg.xi_face)
    {
        if (MT_IS_MAIN_THREAD())
        {
#if !defined(HAVE_MT)
            gui_cfg.endloop = FALSE;
            while (!gui_cfg.endloop && !cfg.stop && !cfg.rbreak && gdk_events_pending())
            {
                int gtrv;

                DEBUG_DEVEL(DBGpass("one gtk_main_iteration()\n"));

                gdk_threads_enter();
                gtrv = gtk_main_iteration();
                gdk_threads_leave();
                if (gtrv)
                {
                    DEBUG_DEVEL(DBGpass("gtk_main_iteration() returned TRUE\n"));
                    break;
                }
            }

            DEBUG_DEVEL(DBGpass("gui_loop_serve() loop aborted due to endloop: %d / stop: %d / break: %d / !gdk_events_pending(): %d\n"),
                        gui_cfg.endloop, cfg.stop, cfg.rbreak, !gdk_events_pending());

            if (gui_cfg._go_bg)
            {
                gui_do_ui_cleanup();
            }

#else

            gui_cfg.endloop = FALSE;
            while (!gui_cfg.endloop && !cfg.stop && !cfg.rbreak && gdk_events_pending())
            {
                int gtrv;

                DEBUG_DEVEL(DBGpass("one gtk_main_iteration()\n"));

                gdk_threads_enter();
                gtrv = gtk_main_iteration();
                gdk_threads_leave();
                if (gtrv)
                {
                    DEBUG_DEVEL(DBGpass("gtk_main_iteration() returned TRUE\n"));
                    break;
                }
            }

            DEBUG_DEVEL(DBGpass("gui_loop_serve() loop aborted due to endloop: %d / stop: %d / break: %d / !gdk_events_pending(): %d\n"),
                        gui_cfg.endloop, cfg.stop, cfg.rbreak, !gdk_events_pending());

#endif
        }
        else
        {
            DEBUG_DEVEL(DBGpass("gui_loop_serve() called from non-MAIN thread. Yielding slightly... endloop: %d / stop: %d / break: %d\n"),
                        gui_cfg.endloop, cfg.stop, cfg.rbreak);

            tl_msleep(2, FALSE);
        }
    }

    DEBUG_PROCE(NULL);
}

#if !defined(HAVE_MT)

void gui_loop_do(void)
{
    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
        gui_cfg.endloop = FALSE;
        while (!gui_cfg.endloop && !cfg.stop && !cfg.rbreak)
        {
            gdk_threads_enter();
            gtk_main_iteration();
            gdk_threads_leave();
        }

  DEBUG_DEVEL(DBGvars("gui_loop_do() loop terminated due to endloop: %d / stop: %d / break: %d\n"), gui_cfg.endloop, cfg.stop, cfg.rbreak);

        if (gui_cfg._go_bg)
        {
            gui_do_ui_cleanup();
        }
    }

    DEBUG_PROCE(NULL);
}

void gui_loop_escape(void)
{
    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
        DEBUG_DEVEL(DBGvars("gui_loop_escape() signals gui to ABORT gui_loop_do/serve GUI message processing\n"));

        gui_cfg.endloop = TRUE;
    }

    DEBUG_PROCE(NULL);
}

#endif /* HAVE_MT */


#ifdef HAVE_MT

void gui_mt_thread_start(int thrnr)
{
    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
        char *ep[7];
        char pom[10];

        sprintf(pom, "%d", thrnr);
        ep[0] = pom;
        ep[1] = "";
        ep[2] = gettext("Starting ...");
        ep[3] = "";
        ep[4] = "";
        ep[5] = "";
        ep[6] = "";
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_clist_append(GTK_CLIST(gui_cfg.status_list), ep);
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }
    }

    DEBUG_PROCE(NULL);
}

void gui_mt_thread_end(int thrnr)
{
    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
            gdk_threads_enter();
        }
        gtk_clist_clear(GTK_CLIST(gui_cfg.status_list));
        if (!MT_IS_MAIN_THREAD())
        {
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
            gdk_flush();        /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
            DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
            gdk_threads_leave();
        }
    }

    DEBUG_PROCE(NULL);
}

#endif


#if !defined(HAVE_MT)

static gint TOut(gpointer data)
{
    DEBUG_PROCS(NULL);

    *(gint *)data = 0;
    gui_loop_escape();

    DEBUG_PROCE(NULL);

    return FALSE;
}

#endif


#if !defined(HAVE_MT)

void gui_msleep(long msec)
{
    gint inid;

    DEBUG_PROCS(NULL);

    inid = gtk_timeout_add(msec, (GtkFunction)TOut, &inid);
    gui_loop_do();
    if (inid)
    {
        gtk_timeout_remove(inid);
    }

    DEBUG_PROCE(NULL);
}

#endif


#if !defined(HAVE_MT)

static void rwCB(gpointer data, gint source, GdkInputCondition condition)
{
    DEBUG_PROCS(NULL);

    gdk_input_remove(*(int *)data);

    *(gint *)data = 0;
    gui_loop_escape();

    DEBUG_PROCE(NULL);
}

#endif


#if !defined(HAVE_MT)

int gui_wait_io(int sock, int for_read, long timeout)
{
    int rv = 0;
    gint iid = 0;
    gint inid = 0;

    DEBUG_PROCS(NULL);

    if (timeout > 0)
    {
        inid = gtk_timeout_add(timeout, (GtkFunction)TOut, &inid);
    }

    iid = gdk_input_add(sock, for_read ? GDK_INPUT_READ : GDK_INPUT_WRITE, (GdkInputFunction)rwCB, &iid);

    gui_loop_do();
    if (inid)
    {
        gtk_timeout_remove(inid);
        inid = 0;
    }
    else if (cfg.conn_timeout > 0)
    {
        errno = ETIMEDOUT;
        rv = -1;
    }

    if (iid)
    {
        gdk_input_remove(iid);
        iid = 0;
    }

    DEBUG_PROCE(NULL);

    return rv;
}

#endif

void gui_xprint(char *str)
{
    char *p;
    bool_t last = 1;
    int ilen;
    char *pline = NULL;
    int row;

    LOCK_GTKLOG;
    if (!MT_IS_MAIN_THREAD())
    {
        DEBUG_DEVEL(DBGvars("### CALLING gdk_threads_enter()\n"));
        gdk_threads_enter();
    }
    gtk_clist_freeze(GTK_CLIST(gui_cfg.logw));
    p = str;
    if (GTK_CLIST(gui_cfg.logw)->rows)
    {
        row = GTK_CLIST(gui_cfg.logw)->rows - 1;
        if (gtk_clist_get_row_data(GTK_CLIST(gui_cfg.logw), row))
        {
            gtk_clist_get_text(GTK_CLIST(gui_cfg.logw), row, 0, &pline);
        }
    }
    while (*p)
    {
        ilen = strcspn(p, "\r\n");
        if (*(p + ilen))
        {
            *(p + ilen) = '\0';
        }
        else
        {
            last = 0;
        }

        if (pline)
        {
            row = GTK_CLIST(gui_cfg.logw)->rows - 1;

            pline = g_strconcat(pline, p, NULL);
            gtk_clist_set_text(GTK_CLIST(gui_cfg.logw), row, 0, pline);
            g_free(pline);
            pline = NULL;

            gtk_clist_set_row_data(GTK_CLIST(gui_cfg.logw), row, cast_int2voidptr(!last));
        }
        else
        {
            row = gtk_clist_append(GTK_CLIST(gui_cfg.logw), &p);

            if (!last)
            {
                gtk_clist_set_row_data(GTK_CLIST(gui_cfg.logw), row, cast_int2voidptr(!last));
            }
        }

        p += ilen + last;
        p += strspn(p, "\r\n");
    }

    if (cfg.xlogsize)
    {
        while (cfg.xlogsize < GTK_CLIST(gui_cfg.logw)->rows)
        {
            gtk_clist_remove(GTK_CLIST(gui_cfg.logw), 0);
        }
    }

    gtk_clist_thaw(GTK_CLIST(gui_cfg.logw));
    if (cfg.log_autoscroll)
    {
    gui_cfg.logvadj->value = ((gui_cfg.logvadj->upper > gui_cfg.logvadj->page_size) 
	? gui_cfg.logvadj->upper - gui_cfg.logvadj->page_size 
	: 0);
        gtk_signal_emit_by_name(GTK_OBJECT(gui_cfg.logvadj), "value_changed");
    }

#if 0                           /* [i_a] experiment while trying to track down the bug that causes the GTK crash :-( */
    {
        GtkCList *cl = GTK_CLIST(gui_cfg.logw);
#define CLIST_UNFROZEN(clist)     (((GtkCList *)(clist))->freeze_count == 0)
#define CLIST_REFRESH(clist) \
    { \
        if (CLIST_UNFROZEN(clist)) \
            GTK_CLIST_GET_CLASS(clist)->refresh((GtkCList *)(clist));\
    }

        if (CLIST_UNFROZEN(cl))
        {
            GTK_CLIST_GET_CLASS(cl)->refresh((GtkCList *)(cl));
        }

        /* Need to refresh the widget --
         * This forces the widget to redraw itself. */
        gtk_widget_queue_draw(GTK_WIDGET(cl));
    }
#endif

    if (!MT_IS_MAIN_THREAD())
    {
        DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave()\n"));
        gdk_flush();            /* See 'Description' at http://developer.gimp.org/api/2.0/gdk/gdk-Threads.html#id3020296 */
        DEBUG_DEVEL(DBGvars("$$$ CALLING gdk_threads_leave() after gdk_flush()\n"));
        gdk_threads_leave();
    }
    UNLOCK_GTKLOG;

#if 0
    gui_loop_serve();
#endif
}


void gui_do_ui_cleanup(void)
{
    DEBUG_PROCS(NULL);

    if (cfg.xi_face)
    {
        if (!MT_IS_MAIN_THREAD())
        {
  DEBUG_DEVEL(DBGvars("gui_do_ui_cleanup() was not called from the main thread; we'll continue on and wait for the main thread to call this function to clean up.\n"));
        }
        else
        {
            cfg.xi_face = FALSE;

#ifdef HAVE_MT
            if (cfg.processing)
            {
                tl_sleep(2);
            }
#endif

            if (gui_cfg.cfg_sch)
            {
                gtk_widget_destroy(gui_cfg.cfg_sch);
            }
            if (gui_cfg.about_shell)
            {
                gtk_widget_destroy(gui_cfg.about_shell);
            }
            if (gui_cfg.scn_load_shell)
            {
                gtk_widget_destroy(gui_cfg.scn_load_shell);
            }
            if (gui_cfg.scn_add_shell)
            {
                gtk_widget_destroy(gui_cfg.scn_add_shell);
            }
            if (gui_cfg.scn_save_shell)
            {
                gtk_widget_destroy(gui_cfg.scn_save_shell);
            }
            if (gui_cfg.cfg_limits)
            {
                gtk_widget_destroy(gui_cfg.cfg_limits);
            }
            if (gui_cfg.config_shell)
            {
                gtk_widget_destroy(gui_cfg.config_shell);
            }
#ifdef WITH_TREE
            if (gui_cfg.tree_shell)
            {
                gtk_widget_destroy(gui_cfg.tree_shell);
            }
#endif

            if (gui_cfg.toplevel)
            {
                gtk_widget_destroy(gui_cfg.toplevel);
            }

            tl_msleep(100, TRUE);

#if !(defined(__CYGWIN__) || defined(WIN32) || defined(__WIN32))
            XCloseDisplay(gdk_display);
#endif

#ifndef HAVE_MT
            dns_server_kill();
#endif
        }
    }

    DEBUG_PROCE(NULL);
}

#endif /* GTK_FACE */

