/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#ifndef _http_h_
#define _http_h_

#include "dllist.h"
#include "bufio.h"
#include "doc.h"
#include "digest_auth.h"


#define DEFAULT_HTTP_PORT       80
#define DEFAULT_HTTP_PROXY_PORT 8080
#define DEFAULT_SSL_PORT        443
#define DEFAULT_SSL_PROXY_PORT  8080

typedef struct
{
    const char *addr;           /* proxy host                        */
    unsigned short port;        /* proxy port                        */
    int penalty;                /* penalty to reject failing proxies */
    unsigned int fails;         /* number of fails for proxy         */
    int is_10;                  /* this is HTTP/1.0 proxy            */
    int ref;                    /* reference counting                */
} http_proxy;

typedef struct
{
    short ver_maj;                    /*** major version of HTTP protocol ***/
    short ver_min;                    /*** minor version of HTTP protocol ***/
    int ret_code;                     /*** HTTP response code             ***/
    const char *text;                 /*** HTTP response code description ***/
} http_response;

typedef struct
{
    bool_t all;
    const char *name;
    const char *val;
} httphdr;

typedef struct
{
    protocol proto;                          /*** protocol ***/
    const char *host;                        /*** hostname ***/
    unsigned short port;                     /*** port     ***/
    bufio *connection;                       /*** socket for persistent connection ***/
    http_digest_info *auth_digest;           /*** HTTP digest access authentification info ***/
    http_digest_info *auth_proxy_digest;     /*** HTTP digest access proxy authentification info ***/
    const char *http_proxy;                  /*** HTTP proxy address ***/
    unsigned short http_proxy_port;          /*** HTTP proxy port ***/
} http_connection;

typedef struct
{
    dllist *form_data;
} http_url_extension;

typedef enum
{
    HTTP_AUTH_NONE = 0,
    HTTP_AUTH_USER = 1,
    HTTP_AUTH_BASIC = 2,
    HTTP_AUTH_DIGEST = 3,
    HTTP_AUTH_NTLM = 4,
    HTTP_AUTH_LAST = 5
} http_auth_type_t;

typedef enum
{
    HTTP_REQ_UNKNOWN,
    HTTP_REQ_GET,
    HTTP_REQ_POST,
    HTTP_REQ_HEAD,
    HTTP_REQ_CONNECT
} http_request_type_t;

typedef struct
{
    const char *name;
    http_auth_type_t id;
} http_auth_type_info_t;

extern char *_md5(const char *data);

extern int http_dummy_proxy_connect(doc *docp, const char *host, int port, const char *proxy_host, int proxy_port);
extern int http_dumy_proxy_connect_real(doc *docp, const char *host, int port, const char *proxy_host, int proxy_port);
extern bufio *http_get_data_socket(doc *docp);
extern int http_repeat_request(doc *docp);
extern int http_get_request(doc *docp);
extern int http_head_request(doc *docp);
extern int http_post_request(doc *docp);
extern int http_read_mime_header(doc *docp, char **buf, ssize_t *len);
extern int http_throw_message_body(doc *docp);
extern void http_handle_redirect(doc *docu, int redir_code);

extern http_response *http_get_response_info(const char *doc_txt);
extern int http_get_response_info2(http_response *dst, const char *doc_txt);
extern void http_response_free(http_response *resp);

extern http_auth_type_t http_get_authorization_type(const char *auth_field);
extern int http_handle_site_auth_info(doc *docp);
extern int http_handle_proxy_auth_info(doc *docp);

extern httphdr *httphdr_parse(const char *str);
extern void httphdr_free(httphdr *hdr);

extern int http_proxy_ref(http_proxy *hp);
extern int http_proxy_unref(http_proxy *hp);
extern void http_proxy_free(http_proxy *hp);
extern http_proxy *http_proxy_parse(const char *pstr);
extern http_proxy *http_proxy_get(void);
extern http_proxy *http_proxy_find(const char *host, int port);
extern void http_proxy_check(http_proxy *hp, doc *docp);

extern const http_auth_type_info_t http_auths[];

#endif

