/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#if defined(HAVE_MOZJS)

#if defined(HAVE_SLIGHTLY_SCREWED_JS_INSTALL) && defined(HAVE_MT) && !defined(JS_THREADSAFE)
#define JS_THREADSAFE  2
#endif

#ifdef HAVE_JSAPI_H
#include <jsapi.h>
#endif

/*
 * But make sure to turn that JS_THREADSAFE #define OFF again
 * before loading any of the other header files -- when we are working
 * with a flaky Linux (SuSe? AMD64?) SpiderMonkey install.
 */
#if defined(HAVE_SLIGHTLY_SCREWED_JS_INSTALL) && defined(HAVE_MT) && (JS_THREADSAFE == 2)
#undef JS_THREADSAFE
#endif

#ifdef HAVE_JSCONFIG_H
#include <jsconfig.h>
#endif
#ifdef HAVE_JSSTDDEF_H
#include <jsstddef.h>
#endif
#ifdef HAVE_JSTYPES_H
#include <jstypes.h>
#endif
#ifdef HAVE_JSARENA_H
#include <jsarena.h>
#endif
#ifdef HAVE_JSUTIL_H
#include <jsutil.h>
#endif
#ifdef HAVE_JSPRF_H
#include <jsprf.h>
#endif
#ifdef HAVE_JSARRAY_H
#include <jsarray.h>
#endif
#ifdef HAVE_JSATOM_H
#include <jsatom.h>
#endif
#ifdef HAVE_JSCNTXT_H
#include <jscntxt.h>
#endif
#ifdef HAVE_JSDBGAPI_H
#include <jsdbgapi.h>
#endif
#ifdef HAVE_JSEMIT_H
#include <jsemit.h>
#endif
#ifdef HAVE_JSFUN_H
#include <jsfun.h>
#endif
#ifdef HAVE_JSGC_H
#include <jsgc.h>
#endif
#ifdef HAVE_JSLOCK_H
#include <jslock.h>
#endif
#ifdef HAVE_JSNUM_H
#include <jsnum.h>
#endif
#ifdef HAVE_JSOBJ_H
#include <jsobj.h>
#endif
#ifdef HAVE_JSPARSE_H
#include <jsparse.h>
#endif
#ifdef HAVE_JSSCOPE_H
#include <jsscope.h>
#endif
#ifdef HAVE_JSSCRIPT_H
#include <jsscript.h>
#endif
#ifdef HAVE_JSOPCODE_H
#include <jsopcode.h>
#endif
#ifdef HAVE_JSPUBTD_H
#include <jspubtd.h>
#endif

#if defined(HAVE_JSDEBUGGER)

#ifdef HAVE_JSDEBUG_H
#include <jsdebug.h>
#endif
#ifdef HAVE_JSDBGAPI_H
#include <jsdbgapi.h>
#endif
#ifdef HAVE_JSDB_H
#include <jsdb.h>
#endif

#endif

#undef   HAVE_JSDEBUGGER


#include "tools.h"
#include "url.h"
#include "dllist.h"
#include "jsbind.h"


#define JSD_LOWLEVEL_SOURCE   1


#define STACK_CHUNK_SIZE 8192



#if !defined(JS_FS_END)
/*
 * Terminating sentinel initializer to put at the end of a JSFunctionSpec array
 * that's passed to JS_DefineFunctions or JS_InitClass.
 */
#define JS_FS_END JS_FS(NULL, NULL, 0, 0, 0)
#endif

#if !defined(JS_FS)
/*
 * Initializer macro for a JSFunctionSpec array element. This is the original
 * kind of native function specifier initializer. Use JS_FN ("fast native", see
 * JSFastNative in jspubtd.h) for all functions that do not need a stack frame
 * when activated.
 */
#define JS_FS(name, call, nargs, flags, extra)                                    \
    { name, call, nargs, flags /* , extra */ }
#endif

#if !defined(JS_OPERATION_WEIGHT_BASE)
#define JS_OPERATION_WEIGHT_BASE 4096
#endif


#if !defined(HAVE_JS_ENCODESTRING)
static char *JS_EncodeString(JSContext *cx, JSString *str)
{
    return JS_GetStringBytes(str);
}
#endif




#define SET_STR(dst_val, cx, src_str)                           \
    if (src_str) \
        *dst_val = STRING_TO_JSVAL(JS_NewStringCopyZ(cx, src_str));\
    else \
        *dst_val = JSVAL_NULL;

static JSRuntime *rt = NULL;
static JSContext *cx = NULL;
static JSObject *go = NULL;
static JSObject *scriptobj = NULL;
static JSScript *script = NULL;
#if defined(HAVE_JSDEBUGGER)
static JSDContext *jsdbg_cx = NULL;
static JSBool jsdbc = JS_FALSE;
#endif
static bool_t pjs_have_pavuk_url_cond_check = FALSE;
static bool_t pjs_have_pavuk_parse_content_check = FALSE;

static JSBool pjs_print(JSContext *, JSObject *, uintN, jsval *, jsval *);
static JSBool pjs_version(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval);
static JSBool pjs_tracing(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval);
static JSBool pjs_load(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval);



/*
 * Add local stub wrapper functions to prevent compiler warnings and run-time failures as reported by MSVC:
 *
 * warning C4232: nonstandard extension used : 'addProperty' : address of dllimport 'JS_PropertyStub' is not static, identity not guaranteed
 */
static JSBool pjs_PropertyStub(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
    return JS_PropertyStub(cx, obj, id, vp);
}

static JSBool pjs_EnumerateStub(JSContext *cx, JSObject *obj)
{
    return JS_EnumerateStub(cx, obj);
}

static JSBool pjs_ResolveStub(JSContext *cx, JSObject *obj, jsval id)
{
    return JS_ResolveStub(cx, obj, id);
}

static JSBool pjs_ConvertStub(JSContext *cx, JSObject *obj, JSType type, jsval *vp)
{
    return JS_ConvertStub(cx, obj, type, vp);
}

static void pjs_FinalizeStub(JSContext *cx, JSObject *obj)
{
    JS_FinalizeStub(cx, obj);
}




/* global class */
static JSClass globalClass = {
    "Global", 0,
    pjs_PropertyStub, pjs_PropertyStub, pjs_PropertyStub, pjs_PropertyStub,
    pjs_EnumerateStub, pjs_ResolveStub, pjs_ConvertStub, pjs_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

/* global methods available */
static JSFunctionSpec globalMethods[] = {
    JS_FS("print", pjs_print, 0, 0, 0),
    JS_FS("load", pjs_load, 1, 0, 0),
    JS_FS("version", pjs_version, 0, 0, 0),
    JS_FS("tracing", pjs_tracing, 0, 0, 0),
    JS_FS_END
};

static void pjs_error(JSContext *cx, const char *message, JSErrorReport *report)
{
    char fmtstr[128];
    char posstr[64];

    if (!report)
    {
        xprintf(1, "pjs: %s\n", message);
        return;
    }

    /* Conditionally ignore reported warnings. */
    if (JSREPORT_IS_WARNING(report->flags) && !cfg.verbose && !cfg.debug)
        return;

    if (report->lineno)
    {
        int colpos = -1;

        if (report->tokenptr && report->linebuf)
        {
            int n = (report->tokenptr - report->linebuf);
            int i;

            /* and now correct the column number for TABs (TAB = 8 chars) */
            for (i = colpos = 0; i < n; i++)
            {
                if (report->linebuf[i] == '\t')
                {
                    /* do as TAB does: align at column mod 8 */
                    colpos += 8;
                    colpos &= ~7;
                }
                else
                {
                    colpos++;
                }
            }
        }

        snprintf(posstr, sizeof(posstr),
            (colpos >= 0 ? " @ line %u, col %u:" : " @ line %u:"),
            report->lineno,
            colpos + 1);
        posstr[64 - 1] = 0;
    }
    else
    {
        posstr[0] = 0;
    }

    /*
     * let's just pray a single line does not run too long, say 100634 bytes. :-(
     * see comments about a buggy glibc elsewhere. (though the bug does not show for
     * a few K's of text on a single line)
     */
    snprintf(fmtstr, sizeof(fmtstr), "pjs:%s%s%s%s: %%.*s\n",
        (report->filename ? " %s" : "%s"),
        (/* report->filename && */ report->lineno ? posstr : ""), /* only show line/col when there's a filename attached? */
        (JSREPORT_IS_WARNING(report->flags) ? " warning" : " error"),
        (JSREPORT_IS_STRICT(report->flags) ? "(strict)" : ""));

    /* embedded newlines -- argh! */
    for (;;)
    {
        int nlpos = strcspn(message, "\n");
        /* print empty lines in the middle of a longer series: */
        if (nlpos != 0 || (message[nlpos] && message[nlpos + 1]))
        {
            xprintf(1, fmtstr,
                (report->filename ? mk_native(report->filename) : ""),
                nlpos,
                message);
            message += nlpos;
        }
        /* if we did not hit an LF or when it's the LF terminating the message: quit */
        if (!message[nlpos] || !message[nlpos + 1])
            break;
        message++;
    }

    if (report->linebuf)
    {
        /*
         * this is probably just a single line, but don't take any risk.
         * Besides, this can be a straight copy&paste of the bit of code
         * just above :-)
         */
        message = report->linebuf;
        /* embedded newlines -- argh! */
        for (;;)
        {
            int nlpos = strcspn(message, "\n");
            /* print empty lines in the middle of a longer series: */
            if (nlpos != 0 || (message[nlpos] && message[nlpos + 1]))
            {
                xprintf(1, fmtstr,
                    (report->filename ? mk_native(report->filename) : ""),
                    nlpos,
                    message);
                message += nlpos;
            }
            /* if we did not hit an LF or when it's the LF terminating the message: quit */
            if (!message[nlpos] || !message[nlpos + 1])
                break;
            message++;
        }
    }

    if (!JSREPORT_IS_WARNING(report->flags))
    {
        LOCK_FAILCNT;

        cfg.stop = TRUE;

#if defined(HAVE_VALUE_JSMSG_OUT_OF_MEMORY)
        if (report->errorNumber != JSMSG_OUT_OF_MEMORY)
        {
            cfg.js_fail_code = 3; /* EXITCODE_OUT_OF_MEMORY */
        }
        else
#endif
        {
            cfg.js_fail_code = 4; /* EXITCODE_RUNTIME_ERROR */
        }

        UNLOCK_FAILCNT;
    }
}


static JSBool pjs_version(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
    JSBool ret = JS_TRUE;

    if (argc > 0)
    {
        int32 ver = 0;
        JSVersion retver;
        ret = JS_ValueToECMAInt32(cx, argv[0], &ver);
        retver = JS_SetVersion(cx, (JSVersion)ver);
        if (rval)
            *rval = INT_TO_JSVAL(retver);
    }
    else
    {
        JSVersion retver;
        retver = JS_GetVersion(cx);
        if (rval)
            *rval = INT_TO_JSVAL(retver);
    }
    return ret;
}


static JSBool pjs_tracing(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
    JSBool ret = JS_TRUE;

#ifdef DEBUG
    int32 argval = 0;

#if JS_THREADED_INTERP
    JS_ReportError(cx, "tracing not supported in JS_THREADED_INTERP builds");
    return JS_FALSE;

#else
    if (argc != 0)
    {
        ret = JS_ValueToECMAInt32(cx, argv[0], &argval);
        if (!ret)
        {
            JS_ReportError(cx, "tracing: illegal argument");
            return ret;
        }

        cx->tracefp = (argval != 0) ? stderr : NULL;
    }
    if (rval)
        *rval = BOOLEAN_TO_JSVAL(cx->tracefp != 0);
#endif
#else /* DEBUG */
    if (rval)
        *rval = JS_FALSE;
#endif /* DEBUG */
    return ret;
}




static JSBool pjs_load_JavaScript_file(JSContext *cx, JSObject *obj, const char *filename)
{
    JSBool rv = JS_FALSE;
    char *buf;
    FILE *file;
    int file_len;
    int rd_len;
    jsval result;

    errno = 0;
    file = fopen(filename, "rb");
    if (!file)
    {
        JS_ReportError(cx, "can't open %s: %s", filename, strerror(errno));
        return JS_FALSE;
    }

    fseek(file, 0, SEEK_END);
    file_len = ftell(file);
    fseek(file, 0, SEEK_SET);

    if (!file_len)
    {
        fclose(file);
        JS_ReportError(cx, "%s is empty", filename);
        return JS_FALSE;
    }

    buf = (char *)malloc(file_len + 1);
    if (!buf)
    {
        fclose(file);
        JS_ReportError(cx, "memory alloc error while trying to read %s", filename);
        return JS_FALSE;
    }
    rd_len = fread(buf, 1, file_len, file);
    fclose(file);

    if (rd_len > 0)
    {
        ASSERT(rd_len <= file_len);
        buf[rd_len] = 0;

        rv = JS_EvaluateScript(cx, obj, buf, file_len, filename, 0, &result);
    }

    free(buf);
    return rv;
}



static JSBool pjs_load(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
    int i;
    JSString *str;
    const char *filename;

    for (i = 0; i < argc; i++)
    {
        str = JS_ValueToString(cx, argv[i]);
        if (!str)
            return JS_FALSE;

        argv[i] = STRING_TO_JSVAL(str);
        filename = JS_GetStringBytes(str);

        if (!pjs_load_JavaScript_file(cx, obj, filename))
            return JS_FALSE;
    }
    return JS_TRUE;
}




static JSBool pjs_print(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rv)
{
    int i;
    JSString *s;

    for (i = 0; i < argc; i++)
    {
        char *str;

        if (!(s = JS_ValueToString(cx, argv[i])) || !(str = JS_EncodeString(cx, s)))
        {
            JS_ReportError(cx, "pjs: print: %s: %d", gettext("bad parameter"), i + 1);
            if (rv)
            {
                *rv = JS_FALSE;
            }
            return JS_FALSE;
        }
        xprintf(1, "%s%s", i ? " " : "", str);
        JS_free(cx, str);
    }

    xprintf(1, "\n");

    if (rv)
    {
        *rv = JS_TRUE;
    }
    return JS_TRUE;
}



/* URL class implementation */
static JSBool pjs_url_get_parent(JSContext *, JSObject *, uintN, jsval *, jsval *);
static JSBool pjs_url_check_cond(JSContext *, JSObject *, uintN, jsval *, jsval *);
static JSBool pjs_url_add_to_queue(JSContext *, JSObject *, uintN, jsval *, jsval *);
static JSBool pjs_url_set_property(JSContext *, JSObject *, jsval, jsval *);
static JSBool pjs_url_get_property(JSContext *, JSObject *, jsval, jsval *);
static void pjs_url_finalize(JSContext *, JSObject *);

static JSClass urlClass = {
    "PavukUrl", JSCLASS_HAS_PRIVATE,
    pjs_PropertyStub, pjs_PropertyStub,
    pjs_url_get_property, pjs_url_set_property,
    pjs_EnumerateStub, pjs_ResolveStub, pjs_ConvertStub,
    pjs_url_finalize,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

#define PJS_IS_URL(v) \
    (JS_GET_CLASS(cx, JSVAL_TO_OBJECT(v)) == &urlClass)

static JSFunctionSpec urlMethods[] = {
    JS_FS("get_parent", pjs_url_get_parent, 1, JSPROP_ENUMERATE, 0),
    JS_FS("check_cond", pjs_url_check_cond, 1, JSPROP_ENUMERATE, 0),
    JS_FS("queue", pjs_url_add_to_queue, 1, JSPROP_ENUMERATE, 0),
    JS_FS_END
};

enum _url_prop
{
    PJSURL_PROTOCOL,
    PJSURL_STATUS,
    PJSURL_LEVEL,
    PJSURL_REF_CNT,
    PJSURL_URLSTR,
    PJSURL_MOVED,

    PJSURL_HTTP_HOST,
    PJSURL_HTTP_PORT,
    PJSURL_HTTP_DOCUMENT,
    PJSURL_HTTP_SEARCHSTR,
    PJSURL_HTTP_ANCHOR_NAME,
    PJSURL_HTTP_USER,
    PJSURL_HTTP_PASS,

    PJSURL_FTP_HOST,
    PJSURL_FTP_PORT,
    PJSURL_FTP_USER,
    PJSURL_FTP_PASS,
    PJSURL_FTP_PATH,
    PJSURL_FTP_ANCHOR_NAME,
    PJSURL_FTP_DIR,

    PJSURL_GOPHER_HOST,
    PJSURL_GOPHER_PORT,
    PJSURL_GOPHER_SELECTOR,

    PJSURL_FILE_FNAME,
    PJSURL_FILE_SEARCHSTR,
    PJSURL_FILE_ANCHOR_NAME,

    PJSURL_UNSUP_URLSTR,

    PJSURL_CHECK_LEVEL,
    PJSURL_MIME_TYPE,
    PJSURL_DOC_SIZE,
    PJSURL_MDTM,
    PJSURL_DOCNR,
    PJSURL_HTML_TAG,
    PJSURL_HTML_DOC,
    PJSURL_HTML_DOC_OFFSET,
    PJSURL_TAG,
    PJSURL_ATTRIB,

    PJSURL_NULL
};

typedef struct
{
    url *urlp;
    cond_info_t *condp;
    int free_url;
} url_priv_t;

static JSPropertySpec urlProperties[] = {
    { "protocol", PJSURL_PROTOCOL, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "status", PJSURL_STATUS, JSPROP_ENUMERATE },
    { "level", PJSURL_LEVEL, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "ref_cnt", PJSURL_REF_CNT, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "urlstr", PJSURL_URLSTR, JSPROP_ENUMERATE | JSPROP_READONLY },
    { NULL, 0, 0 },
};

static url_priv_t *pjs_url_obj_set_url(JSContext *cx, JSObject *obj, url *purl, cond_info_t *condp)
{
    url_priv_t *up;

    up = _malloc(sizeof(*up));

    up->urlp = purl;
    up->condp = condp;
    up->free_url = FALSE;

    JS_SetPrivate(cx, obj, up);

#define NEW_PROP(name, id) \
    JS_DefinePropertyWithTinyId(cx, obj, name, id, JSVAL_VOID, \
                                pjs_url_get_property, pjs_PropertyStub, \
                                JSPROP_ENUMERATE | JSPROP_READONLY);

    if (purl->moved_to)
        NEW_PROP("moved_url", PJSURL_MOVED);

    switch (purl->type)
    {
    case URLT_HTTP:
    case URLT_HTTPS:
        NEW_PROP("http_host", PJSURL_HTTP_HOST);
        NEW_PROP("http_port", PJSURL_HTTP_PORT);
        NEW_PROP("http_document", PJSURL_HTTP_DOCUMENT);
        NEW_PROP("http_searchstr", PJSURL_HTTP_SEARCHSTR);
        NEW_PROP("http_anchor_name", PJSURL_HTTP_ANCHOR_NAME);
        NEW_PROP("http_user", PJSURL_HTTP_USER);
        NEW_PROP("http_password", PJSURL_HTTP_PASS);
        break;

    case URLT_FTP:
    case URLT_FTPS:
        NEW_PROP("ftp_host", PJSURL_FTP_HOST);
        NEW_PROP("ftp_port", PJSURL_FTP_PORT);
        NEW_PROP("ftp_user", PJSURL_FTP_USER);
        NEW_PROP("ftp_password", PJSURL_FTP_PASS);
        NEW_PROP("ftp_path", PJSURL_FTP_PATH);
        NEW_PROP("ftp_anchor_name", PJSURL_FTP_ANCHOR_NAME);
        NEW_PROP("ftp_dir", PJSURL_FTP_DIR);
        break;

    case URLT_FILE:
        NEW_PROP("file_name", PJSURL_FILE_FNAME);
        NEW_PROP("file_searchstr", PJSURL_FILE_SEARCHSTR);
        NEW_PROP("file_anchor_name", PJSURL_FILE_ANCHOR_NAME);
        break;

    case URLT_GOPHER:
        NEW_PROP("gopher_host", PJSURL_GOPHER_HOST);
        NEW_PROP("gopher_port", PJSURL_GOPHER_PORT);
        NEW_PROP("gopher_selector", PJSURL_GOPHER_SELECTOR);
        break;

    case URLT_UNKNOWN:
        NEW_PROP("unsupported_urlstr", PJSURL_UNSUP_URLSTR);

    default:
        break;
    }

    if (up->condp)
    {
        NEW_PROP("check_level", PJSURL_CHECK_LEVEL);
        if (up->condp->mimet)
            NEW_PROP("mime_type", PJSURL_MIME_TYPE);
        if (up->condp->full_tag)
            NEW_PROP("html_tag", PJSURL_HTML_TAG);
        if (up->condp->size >= 0)
            NEW_PROP("doc_size", PJSURL_DOC_SIZE);
        if (up->condp->time)
            NEW_PROP("modification_time", PJSURL_MDTM);
        if (up->condp->urlnr)
            NEW_PROP("doc_number", PJSURL_DOCNR);
        if (up->condp->tag)
            NEW_PROP("tag", PJSURL_TAG);
        if (up->condp->attrib)
            NEW_PROP("attrib", PJSURL_ATTRIB);
        if (up->condp->html_doc)
        {
            NEW_PROP("html_doc", PJSURL_HTML_DOC);
            NEW_PROP("html_doc_offset", PJSURL_HTML_DOC_OFFSET);
        }
    }

    return up;
}

/* constructor */
static JSBool pjs_url_new(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rv)
{
    url_priv_t *up;
    url *urlp = NULL;

    if (argc != 1 || !JSVAL_IS_STRING(argv[0]))
    {
        JS_ReportError(cx, gettext("PavukUrl constructor failed\n"));
        return JS_FALSE;
    }

    urlp = url_parse(JS_GetStringBytes(JSVAL_TO_STRING(argv[0])));
    if (urlp->type == URLT_FROMPARENT)
    {
        JS_ReportError(cx, gettext("PavukUrl constructor does not accept parented URLs\n"));
        return JS_FALSE;
    }
    up = pjs_url_obj_set_url(cx, obj, urlp, NULL);
    up->free_url = TRUE;

    return JS_TRUE;
}



/* destructor */
static void pjs_url_finalize(JSContext *cx, JSObject *obj)
{
    url_priv_t *up;

    up = JS_GetPrivate(cx, obj);

    if (up)
    {
        if (up->free_url)
        {
            free_deep_url(up->urlp);
            _free(up->urlp);
        }
        _free(up);
    }
}

static JSObject *pjs_url_to_jsobj(JSContext *cx, url *purl, cond_info_t *condp)
{
    JSObject *obj;

    obj = JS_NewObject(cx, &urlClass, NULL, NULL);
    pjs_url_obj_set_url(cx, obj, purl, condp);

    return obj;
}

static JSBool pjs_url_get_property(JSContext *cx, JSObject *obj, jsval id, jsval *val)
{
    if (JSVAL_IS_INT(id))
    {
        url_priv_t *up;
        int n;
        const char *p;

        up = JS_GetPrivate(cx, obj);

        switch (JSVAL_TO_INT(id))
        {
        case PJSURL_PROTOCOL:
            if (up->urlp->type >= 0 && up->urlp->type < prottable_size())
            {
                p = prottable[up->urlp->type].urlid;
            }
            else
            {
                p = prottable[URLT_UNKNOWN].urlid;
            }
            SET_STR(val, cx, p);
            break;

        case PJSURL_STATUS:
            *val = INT_TO_JSVAL(up->urlp->status);
            break;

        case PJSURL_LEVEL:
            *val = INT_TO_JSVAL(up->urlp->level);
            break;

        case PJSURL_REF_CNT:
            LOCK_URL(up->urlp);
            *val = INT_TO_JSVAL(up->urlp->ref_cnt);
            UNLOCK_URL(up->urlp);
            break;

        case PJSURL_URLSTR:
            p = url_to_urlstr(up->urlp, TRUE);
            SET_STR(val, cx, p);
            _free(p);
            break;

        case PJSURL_HTTP_HOST:
        case PJSURL_FTP_HOST:
        case PJSURL_GOPHER_HOST:
            p = url_get_site(up->urlp);
            SET_STR(val, cx, p);
            break;

        case PJSURL_HTTP_PORT:
        case PJSURL_FTP_PORT:
        case PJSURL_GOPHER_PORT:
            n = url_get_port(up->urlp);
            *val = INT_TO_JSVAL(n);
            break;

        case PJSURL_HTTP_DOCUMENT:
        case PJSURL_FTP_PATH:
        case PJSURL_GOPHER_SELECTOR:
        case PJSURL_FILE_FNAME:
            p = url_get_path(up->urlp);
            SET_STR(val, cx, p);
            break;

        case PJSURL_HTTP_ANCHOR_NAME:
        case PJSURL_FTP_ANCHOR_NAME:
        case PJSURL_FILE_ANCHOR_NAME:
            p = url_get_anchor_name(up->urlp);
            SET_STR(val, cx, p);
            break;

        case PJSURL_HTTP_SEARCHSTR:
        case PJSURL_FILE_SEARCHSTR:
            p = url_get_search_str(up->urlp);
            SET_STR(val, cx, p);
            break;

        case PJSURL_HTTP_USER:
        case PJSURL_FTP_USER:
            p = url_get_user(up->urlp, NULL);
            SET_STR(val, cx, p);
            break;

        case PJSURL_HTTP_PASS:
        case PJSURL_FTP_PASS:
            p = url_get_pass(up->urlp, NULL);
            SET_STR(val, cx, p);
            break;

        case PJSURL_FTP_DIR:
            *val = up->urlp->p.ftp.dir ? JS_TRUE : JS_FALSE;
            break;

        case PJSURL_UNSUP_URLSTR:
            p = up->urlp->urlstr;
            SET_STR(val, cx, p);
            break;

        case PJSURL_CHECK_LEVEL:
            ASSERT(up->condp);
            *val = INT_TO_JSVAL(up->condp->level);
            break;

        case PJSURL_MIME_TYPE:
            ASSERT(up->condp);
            SET_STR(val, cx, up->condp->mimet);
            break;

        case PJSURL_DOC_SIZE:
            ASSERT(up->condp);
            *val = INT_TO_JSVAL(up->condp->size);
            break;

        case PJSURL_MDTM:
            ASSERT(up->condp);
            *val = INT_TO_JSVAL(up->condp->time);
            break;

        case PJSURL_DOCNR:
            ASSERT(up->condp);
            *val = INT_TO_JSVAL(up->condp->urlnr);
            break;

        case PJSURL_HTML_TAG:
            ASSERT(up->condp);
            SET_STR(val, cx, up->condp->full_tag);
            break;

        case PJSURL_TAG:
            ASSERT(up->condp);
            SET_STR(val, cx, up->condp->tag);
            break;

        case PJSURL_ATTRIB:
            ASSERT(up->condp);
            SET_STR(val, cx, up->condp->attrib);
            break;

        case PJSURL_HTML_DOC:
            ASSERT(up->condp);
            SET_STR(val, cx, up->condp->html_doc);
            break;

        case PJSURL_HTML_DOC_OFFSET:
            ASSERT(up->condp);
            *val = INT_TO_JSVAL(up->condp->html_doc_offset);
            break;

        case PJSURL_MOVED:
            if (up->urlp->moved_to)
            {
                JSObject *robj;
                robj = pjs_url_to_jsobj(cx, up->urlp->moved_to, NULL);
                *val = OBJECT_TO_JSVAL(robj);
            }
            else
                *val = JSVAL_NULL;
            break;
        }
    }

    return JS_TRUE;
}

static JSBool pjs_url_set_property(JSContext *cx, JSObject *obj, jsval id, jsval *val)
{
    if (JSVAL_IS_INT(id))
    {
        url_priv_t *up;
        int32 status;

        up = JS_GetPrivate(cx, obj);

        switch (JSVAL_TO_INT(id))
        {
        case PJSURL_STATUS:
            if (!JS_ValueToInt32(cx, *val, &status))
            {
                JS_ReportError(cx, "PavukUrl.status: %s\n", gettext("invalid assignment"));
                return JS_FALSE;
            }
            up->urlp->status = status;
            break;
        }
    }

    return JS_TRUE;
}

static JSBool pjs_url_get_parent(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rv)
{
    int n;
    url *rurl;
    url_priv_t *up;
    dllist *ptr;
    JSObject *robj;

    if (argc != 1 || !JSVAL_IS_INT(argv[0]))
    {
        JS_ReportError(cx, "PavukUrl.get_parent: %s\n", gettext("bad parent index"));
        return JS_FALSE;
    }

    n = JSVAL_TO_INT(argv[0]);
    up = JS_GetPrivate(cx, obj);

    ptr = dllist_nth(up->urlp->parent_url, n);

    if (ptr)
    {
        rurl = (url *)ptr->data;

        robj = pjs_url_to_jsobj(cx, rurl, NULL);

        *rv = OBJECT_TO_JSVAL(robj);
    }
    else
        *rv = JSVAL_NULL;

    return JS_TRUE;
}


static JSBool pjs_url_add_to_queue(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rv)
{
    JS_ReportError(cx, "PavukUrl.queue: %s\n", gettext("unsupported command"));
    *rv = JSVAL_FALSE;
    return JS_FALSE;
}


static JSBool pjs_url_check_cond(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rv)
{
    url_priv_t *up;
    char *name;
    JSString *s;
    int r, i;
    cond_info_t condp = { 0 };

    up = JS_GetPrivate(cx, obj);

    if (up->condp)
    {
        memcpy(&condp, up->condp, sizeof(condp));
        /*
         * do NOT copy params; these are 'reason'-dependent, and
         * the JS coder specifies his own 'reason' code.
         *
         * Besides, we don't want the original params all freed
         * at the end here as that will definitely cause double
         * free() attempts once we return from all this!
         */
        condp.params = NULL;
        condp.reason = CONDT_UNSUP;
    }

    if (argc < 1)
    {
        JS_ReportError(cx, "PavukUrl.check_cond: %s\n", gettext("not enough parameters"));
        *rv = JSVAL_FALSE;
        return JS_FALSE;
    }

    s = JS_ValueToString(cx, argv[0]);
    name = JS_GetStringBytes(s);

    for (i = 1; i < argc; i++)
    {
        char *p;

        s = JS_ValueToString(cx, argv[i]);
        p = JS_GetStringBytes(s);

        if (p)
        {
            condp.params = dllist_append(condp.params, p);
        }
    }

    r = url_append_one_condition(name, up->urlp, &condp);

    dllist_free_all(condp.params);

    if (r < 0)
    {
        JS_ReportError(cx, "pjs: PavukUrl.check_cond: %s - \"%s\"", gettext("unknown limiting condition"), name);

        *rv = JSVAL_FALSE;      /* reject / fail! */
        return JS_FALSE;
    }

    *rv = r ? JSVAL_TRUE : JSVAL_FALSE;

    return JS_TRUE;
}

static JSObject *pjs_url_class_init(JSContext *cx, JSObject *obj)
{
    JSObject *prot;

    prot = JS_InitClass(cx, obj, NULL, &urlClass, pjs_url_new, 0, urlProperties, urlMethods, NULL, NULL);
    if (!prot)
        return NULL;

    /*
     * Null Call.prototype's proto slot so that Object.prototype.* does not
     * pollute the scope of heavyweight functions.
     */
    OBJ_SET_PROTO(cx, prot, NULL);

    return prot;
}

#if 0
/* PavukDocument class implementation */
static JSObject *pjs_doc_to_jsobj(JSContext *cx, doc *docp)
{
    JSObject *obj;

    obj = JS_NewObject(cx, &docClass, NULL, NULL);
    pjs_doc_obj_set_doc(cx, obj, docp);

    return obj;
}
#endif


/* PavukFnrules class implementation */
static JSBool pjs_fnrules_get_macro(JSContext *, JSObject *, uintN, jsval *, jsval *);
static JSBool pjs_fnrules_get_sub(JSContext *, JSObject *, uintN, jsval *, jsval *);
static JSBool pjs_fnrules_get_property(JSContext *, JSObject *, jsval, jsval *);

static JSClass fnrulesClass = {
    "PavukFnrules", JSCLASS_HAS_PRIVATE,
    pjs_PropertyStub, pjs_PropertyStub,
    pjs_fnrules_get_property, pjs_PropertyStub,
    pjs_EnumerateStub, pjs_ResolveStub, pjs_ConvertStub, pjs_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSFunctionSpec fnrulesMethods[] = {
    JS_FS("get_macro", pjs_fnrules_get_macro, 1, JSPROP_ENUMERATE, 0),
    JS_FS("get_sub", pjs_fnrules_get_sub, 1, JSPROP_ENUMERATE, 0),
    JS_FS_END
};

enum _fnrules_prop
{
    PJSFNRULES_URL,
    PJSFNRULES_PATTERN,
    PJSFNRULES_LISP_PATTERN,
    PJSFNRULES_PATTERN_TYPE,
    PJSFNRULES_URLSTR,
    PJSFNRULES_SCHEME,
    PJSFNRULES_PASSWD,
    PJSFNRULES_USER,
    PJSFNRULES_HOST,
    PJSFNRULES_DOMAIN,
    PJSFNRULES_PORT,
    PJSFNRULES_PATH,
    PJSFNRULES_NAME,
    PJSFNRULES_BASENAME,
    PJSFNRULES_EXTENSION,
    PJSFNRULES_QUERY,
    PJSFNRULES_POST_QUERY,
    PJSFNRULES_DEFLT,
    PJSFNRULES_MIME_TYPE,
    PJSFNRULES_BASIC_MIME_TYPE,
    PJSFNRULES_MIME_TYPE_ATTRIBS,
    PJSFNRULES_MIME_TYPE_EXT,

    PJSFNRULES_NULL
};

static JSPropertySpec fnrulesProperties[] =
{
    { "url", PJSFNRULES_URL, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "pattern", PJSFNRULES_PATTERN, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "lisp_pattern", PJSFNRULES_LISP_PATTERN, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "pattern_type", PJSFNRULES_PATTERN_TYPE, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "urlstr", PJSFNRULES_URLSTR, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "scheme", PJSFNRULES_SCHEME, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "passwd", PJSFNRULES_PASSWD, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "user", PJSFNRULES_USER, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "host", PJSFNRULES_HOST, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "domain", PJSFNRULES_DOMAIN, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "port", PJSFNRULES_PORT, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "path", PJSFNRULES_PATH, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "name", PJSFNRULES_NAME, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "basename", PJSFNRULES_BASENAME, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "extension", PJSFNRULES_EXTENSION, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "query", PJSFNRULES_QUERY, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "post_query", PJSFNRULES_POST_QUERY, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "deflt", PJSFNRULES_DEFLT, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "mime_type", PJSFNRULES_MIME_TYPE, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "basic_mime_type", PJSFNRULES_BASIC_MIME_TYPE, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "mime_type_attribs", PJSFNRULES_MIME_TYPE_ATTRIBS, JSPROP_ENUMERATE | JSPROP_READONLY },
    { "mime_type_ext", PJSFNRULES_MIME_TYPE_EXT, JSPROP_ENUMERATE | JSPROP_READONLY },

    { NULL, 0, 0 },
};

static JSObject *pjs_fnrules_to_jsobj(JSContext *cx, struct lfname_lsp_interp *interp)
{
    JSObject *obj;

    obj = JS_NewObject(cx, &fnrulesClass, NULL, NULL);
    JS_SetPrivate(cx, obj, interp);

    return obj;
}

/* constructor */
static JSBool pjs_fnrules_new(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rv)
{
    JS_ReportError(cx, gettext("PavukFnrules constructor call prohibited from script\n"));
    return JS_TRUE;
}

static JSBool pjs_fnrules_get_property(JSContext *cx, JSObject *obj, jsval id, jsval *val)
{
    if (JSVAL_IS_INT(id))
    {
        struct lfname_lsp_interp *interp;
        JSObject *uobj;

        interp = JS_GetPrivate(cx, obj);

        switch (JSVAL_TO_INT(id))
        {
        case PJSFNRULES_URL:
            uobj = pjs_url_to_jsobj(cx, interp->urlp, NULL);
            *val = OBJECT_TO_JSVAL(uobj);
            break;

        case PJSFNRULES_PATTERN:
            SET_STR(val, cx, interp->orig->matchstr);
            break;

        case PJSFNRULES_LISP_PATTERN:
            SET_STR(val, cx, interp->orig->transstr);
            break;

        case PJSFNRULES_PATTERN_TYPE:
            *val = INT_TO_JSVAL(interp->orig->type);
            break;

        case PJSFNRULES_URLSTR:
            SET_STR(val, cx, interp->urlstr);
            break;

        case PJSFNRULES_SCHEME:
            SET_STR(val, cx, interp->scheme);
            break;

        case PJSFNRULES_PASSWD:
            SET_STR(val, cx, interp->passwd);
            break;

        case PJSFNRULES_USER:
            SET_STR(val, cx, interp->user);
            break;

        case PJSFNRULES_HOST:
            SET_STR(val, cx, interp->host);
            break;

        case PJSFNRULES_DOMAIN:
            SET_STR(val, cx, interp->domain);
            break;

        case PJSFNRULES_PORT:
            SET_STR(val, cx, interp->port);
            break;

        case PJSFNRULES_PATH:
            SET_STR(val, cx, interp->path);
            break;

        case PJSFNRULES_NAME:
            SET_STR(val, cx, interp->name);
            break;

        case PJSFNRULES_BASENAME:
            SET_STR(val, cx, interp->basename);
            break;

        case PJSFNRULES_EXTENSION:
            SET_STR(val, cx, interp->extension);
            break;

        case PJSFNRULES_QUERY:
            SET_STR(val, cx, interp->query);
            break;

        case PJSFNRULES_POST_QUERY:
            SET_STR(val, cx, interp->post_query);
            break;

        case PJSFNRULES_DEFLT:
            SET_STR(val, cx, interp->deflt);
            break;

        case PJSFNRULES_MIME_TYPE:
            SET_STR(val, cx, interp->mime_type);
            break;

        case PJSFNRULES_BASIC_MIME_TYPE:
            SET_STR(val, cx, interp->basic_mime_type);
            break;

        case PJSFNRULES_MIME_TYPE_ATTRIBS:
            SET_STR(val, cx, interp->mime_type_attribs);
            break;

        case PJSFNRULES_MIME_TYPE_EXT:
            SET_STR(val, cx, interp->mime_type_ext);
            break;

        default:
            return JS_FALSE;
        }
    }
    return JS_TRUE;
}

static JSBool pjs_fnrules_get_macro(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *val)
{
    struct lfname_lsp_interp *interp;
    const char *v;
    JSString *s;

    if (argc != 1 || !(s = JS_ValueToString(cx, argv[0])))
    {
        JS_ReportError(cx, "PavukFnrules.get_macro: %s\n", gettext("bad parameters"));
        return JS_FALSE;
    }

    interp = JS_GetPrivate(cx, obj);
    v = JS_GetStringBytes(s);

    if (v && v[0] == '%' && lfname_check_macro(v[1]) && !v[2])
    {
        v = lfname_interp_get_macro(interp, v[1]);
    }
    SET_STR(val, cx, v);

    return JS_TRUE;
}

static JSBool pjs_fnrules_get_sub(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *val)
{
    struct lfname_lsp_interp *interp;
    char *v;
    int32 n;

    if (argc != 1 || !JSVAL_IS_INT(argv[0]))
    {
        JS_ReportError(cx, "PavukFnrules.get_sub: %s\n", gettext("bad parameters"));
        return JS_FALSE;
    }

    interp = JS_GetPrivate(cx, obj);

    JS_ValueToInt32(cx, argv[0], &n);

    v = lfname_re_sub(interp->orig, interp->urlstr, n);

    SET_STR(val, cx, v);

    return JS_TRUE;
}

static JSObject *pjs_fnrules_class_init(JSContext *cx, JSObject *obj)
{
    JSObject *prot;

    prot = JS_InitClass(cx, obj, NULL, &fnrulesClass, pjs_fnrules_new, 0, fnrulesProperties, fnrulesMethods, NULL, NULL);
    if (!prot)
        return NULL;

    /*
     * Null Call.prototype's proto slot so that Object.prototype.* does not
     * pollute the scope of heavyweight functions.
     */
    OBJ_SET_PROTO(cx, prot, NULL);

    return prot;
}

static int pjs_func_exist(JSContext *cx, JSObject *obj, char *name)
{
    jsval rv;

    if (!JS_GetProperty(cx, obj, name, &rv) || rv == JSVAL_VOID)
        return FALSE;
    else
        return TRUE;
}


/* this bit of code is derived from the latest Mozilla SpiderMonkey js.c test app code */


static int psj_BranchCount = 0;
static int psj_BranchLimit = 256;


#if defined(HAVE_JS_SETOPERATIONCALLBACK)

static JSBool psj_BranchCallback(JSContext *cx)
{
    if (++psj_BranchCount >= psj_BranchLimit)
    {
        JSScript *script = (cx->fp ? cx->fp->script : NULL);

        if (script)
        {
            xprintf(1, "psj: %s:%u: script branch callback (%u callbacks)\n",
                (script->filename ? mk_native(script->filename) : "-"),
                script->lineno,
                psj_BranchLimit);
        }
        else
        {
            xprintf(1, "psj: native branch callback (%u callbacks)\n",
                    psj_BranchLimit);
        }
        psj_BranchCount = 0;

        return JS_FALSE;
    }

    if (!(psj_BranchCount & 0xff))
    {
        if (!(psj_BranchCount & 0x3fff))
        {
            JS_MaybeGC(cx);
        }
        else
        {
#if defined(HAVE_JS_YIELDREQUEST)
            JS_YieldRequest(cx);
#endif
        }
    }
    return JS_TRUE;
}

#endif /* defined(HAVE_JS_SETOPERATIONCALLBACK) */


static JSBool psj_ContextNewSetup(JSContext *cx)
{
    psj_BranchCount = 0;
    psj_BranchLimit = 0x40000;

    JS_SetErrorReporter(cx, pjs_error);
#if defined(HAVE_JS_SETVERSION) && defined(HAVE_VALUE_JSVERSION_LATEST)
    JS_SetVersion(cx, JSVERSION_LATEST);
#elif defined(HAVE_JS_SETVERSION) && defined(JSVERSION)
    JS_SetVersion(cx, JSVERSION);
#endif
#if defined(HAVE_JS_SETCONTEXTTHREAD)
    JS_SetContextThread(cx);
#endif

    /*
     * Disable checking for stack overflow if limit is zero.
     */
    JS_SetThreadStackLimit(cx, 0);

#if defined(HAVE_JS_SETSCRIPTSTACKQUOTA)
    JS_SetScriptStackQuota(cx, TL_MIN(JS_DEFAULT_SCRIPT_STACK_QUOTA, 1000000L));
#endif

#if defined(HAVE_JS_SETOPERATIONCALLBACK)
    JS_SetOperationCallback(cx, psj_BranchCallback, JS_OPERATION_WEIGHT_BASE);
    JS_ToggleOptions(cx, JSOPTION_NATIVE_BRANCH_CALLBACK);
#endif

    return JS_TRUE;
}




#if defined(HAVE_JS_SETCONTEXTCALLBACK)

static JSBool psj_ContextCallback(JSContext *cx, uintN contextOp)
{
    JSBool retv = JS_TRUE;

#if defined(HAVE_VALUE_JSCONTEXT_NEW)
    if (contextOp == JSCONTEXT_NEW)
    {
        retv = psj_ContextNewSetup(cx);
    }
#endif /* defined(HAVE_VALUE_JSCONTEXT_NEW) */

    return JS_TRUE;
}

#endif /* defined(HAVE_JS_SETCONTEXTCALLBACK) */



#if defined(HAVE_JSDEBUGGER)
#if defined(JSD_LOWLEVEL_SOURCE)

/*
 * This facilitates sending source to JSD (the debugger system) in the shell
 * where the source is loaded using the JSFILE hack in jsscan. The function
 * below is used as a callback for the jsdbgapi JS_SetSourceHandler hook.
 * A more normal embedding (e.g. mozilla) loads source itself and can send
 * source directly to JSD without using this hook scheme.
 */
static void SendSourceToJSDebugger(const char *filename, uintN lineno,
        jschar *str, size_t length, void **listenerTSData, JSDContext *jsdc)
{
    JSDSourceText *jsdsrc = (JSDSourceText *)*listenerTSData;

    if (!jsdsrc)
    {
        if (!filename)
        {
            filename = "typein";
        }

        if (1 == lineno)
        {
            jsdsrc = JSD_NewSourceText(jsdc, filename);
        }
        else
        {
            jsdsrc = JSD_FindSourceForURL(jsdc, filename);
            if (jsdsrc && JSD_SOURCE_PARTIAL != JSD_GetSourceStatus(jsdc, jsdsrc))
            {
                jsdsrc = NULL;
            }
        }
    }
    if (jsdsrc)
    {
        jsdsrc = JSD_AppendUCSourceText(jsdc, jsdsrc, str, length, JSD_SOURCE_PARTIAL);
    }

    *listenerTSData = jsdsrc;
}

#endif /* JSD_LOWLEVEL_SOURCE */
#endif



int pjs_init(void)
{
    DEBUG_JS("pjs_init()\n");

    LOCK_MOZJS;

    if (!rt)
    {
        rt = JS_Init(1000000L);
        if (!rt)
            return 1;

#if defined(HAVE_JS_SETCONTEXTCALLBACK)
        (void)JS_SetContextCallback(rt, psj_ContextCallback);
#endif
        cx = JS_NewContext(rt, STACK_CHUNK_SIZE);
        if (!cx)
            return 1;

#if !defined(HAVE_JS_SETCONTEXTCALLBACK) || !defined(HAVE_VALUE_JSCONTEXT_NEW)
        /* this code should only be included if the psj_ContextCallback function above could NOT be registered. */
        if (psj_ContextNewSetup(cx) != JS_TRUE)
            return 1;

#endif  /* !defined(HAVE_JS_SETCONTEXTCALLBACK) || !defined(HAVE_VALUE_JSCONTEXT_NEW) */

        /* must be called here, or your multithread-support-compiled JS code will barf at arbitrary times in the next lines */
#if defined(HAVE_JS_BEGINREQUEST)
        JS_BeginRequest(cx);
#endif

        go = JS_NewObject(cx, &globalClass, NULL, NULL);
        if (!go)
            return 1;

        if (!JS_InitStandardClasses(cx, go))
            return 1;

        if (!JS_DefineFunctions(cx, go, globalMethods))
            return 1;

#if 0
        it = JS_DefineObject(cx, glob, "it", &its_class, NULL, 0);
        if (!it)
            return 1;

        if (!JS_DefineProperties(cx, it, its_props))
            return 1;

        if (!JS_DefineFunctions(cx, it, its_methods))
            return 1;

#endif

#if defined(HAVE_JSDEBUGGER)
        /*
         * XXX A command line option to enable debugging (or not) would be good
         */
        jsdbg_cx = JSD_DebuggerOnForUser(rt, NULL, NULL);
        if (!jsdbg_cx)
            return 1;

        JSD_JSContextInUse(jsdbg_cx, cx);
#if defined(JSD_LOWLEVEL_SOURCE)
        JS_SetSourceHandler(rt, SendSourceToJSDebugger, jsdbg_cx);
#endif  /* JSD_LOWLEVEL_SOURCE */

        jsdbc = JSDB_InitDebugger(rt, jsdbg_cx, 0);
#endif  /* JSDEBUGGER */


        if (!pjs_url_class_init(cx, go))
            return 1;

        if (!pjs_fnrules_class_init(cx, go))
            return 1;

        pjs_load_script(priv_cfg.js_script_file);

        pjs_have_pavuk_url_cond_check = pjs_func_exist(cx, go, "pavuk_url_cond_check");
        pjs_have_pavuk_parse_content_check = pjs_func_exist(cx, go, "pavuk_parse_content");
    }

    UNLOCK_MOZJS;

    return 0;
}

int pjs_destroy(void)
{
    DEBUG_JS("pjs_destroy()\n");

    LOCK_MOZJS;

    if (rt)
    {
        if (script)
        {
            JS_RemoveRoot(cx, &scriptobj);
            JS_DestroyScript(cx, script);
        }
        JS_GC(cx);

#if defined(HAVE_JSDEBUGGER)
        if (jsdbg_cx)
        {
            if (jsdbc)
            {
                JSDB_TermDebugger(jsdbg_cx);
            }
            JSD_DebuggerOff(jsdbg_cx);
        }
#endif  /* JSDEBUGGER */

#if defined(HAVE_JS_ENDREQUEST)
        JS_EndRequest(cx);      /* thread safety code */
#endif

        JS_DestroyContext(cx);
        JS_Finish(rt);          /* == JS_DestroyRuntime(rt); */
        JS_ShutDown();

#if defined(HAVE_JSDEBUGGER)
        jsdbg_cx = NULL;
        jsdbc = JS_FALSE;
#endif
        cx = NULL;
        rt = NULL;
        go = NULL;
        script = NULL;
        scriptobj = NULL;
        pjs_have_pavuk_url_cond_check = FALSE;
        pjs_have_pavuk_parse_content_check = FALSE;
    }

    UNLOCK_MOZJS;

    return 0;
}

int pjs_load_script(const char *name)
{
    jsval rv;

    DEBUG_JS("pjs_load_script(%s)\n", name);

    JS_ClearPendingException(cx);
    if (script)
    {
        JS_RemoveRoot(cx, &scriptobj);
        JS_DestroyScript(cx, script);
        script = NULL;
        scriptobj = NULL;
    }

    if (name)
    {
        script = JS_CompileFile(cx, go, name);

        if (!script)
            return -1;

        scriptobj = JS_NewScriptObject(cx, script);
        JS_AddNamedRoot(cx, &scriptobj, "pjs_script");

        JS_ExecuteScript(cx, go, script, &rv);
    }

    return 0;
}

int pjs_load_script_string(const char *str)
{
    jsval rv;

    DEBUG_JS("pjs_load_script_string(%s)\n", str);

    LOCK_MOZJS;
    JS_ClearPendingException(cx);
    if (script)
    {
        JS_RemoveRoot(cx, &scriptobj);
        JS_DestroyScript(cx, script);
        script = NULL;
        scriptobj = NULL;
    }

    script = JS_CompileScript(cx, go, str, strlen(str), "intbuffer", 0);
    if (!script)
    {
        UNLOCK_MOZJS;
        return -1;
    }

    scriptobj = JS_NewScriptObject(cx, script);
    JS_AddNamedRoot(cx, &scriptobj, "pjs_script");

    JS_ExecuteScript(cx, go, script, &rv);

    UNLOCK_MOZJS;
    return 0;
}

int pjs_execute(const char *scr)
{
    jsval rval;
    JSBool rv;

    DEBUG_JS("pjs_execute(%s)\n", scr);

    LOCK_MOZJS;
    rv = JS_EvaluateScript(cx, go, scr, strlen(scr), "pjs", 0, &rval);
    UNLOCK_MOZJS;

    return rv;
}

/*
 * Execute the JavaScript function 'pavuk_url_cond_check(URL, level)'
 * for the given URL 'urlp' and accompanying condition parameters
 * in 'condp'.
 *
 * Return -1 when an unsupported situation is found, for example
 * when the JavaScript function does not exist in the provided
 * JavaScript file (loaded elsewhere using
 *   cfg.js_script_file
 *
 * Otherwise return an boolean/integer value as returned by the
 * JavaScript function.
 */
int pjs_run_cond_check_func(url *urlp, cond_info_t *condp)
{
    jsval param[2];
    jsval rv;
    JSObject *ourl;
    int ret = -1;

    DEBUG_JS("pjs_run_cond_check_func(...)\n");

    if (!pjs_have_pavuk_url_cond_check || !script)
    {
        return -1;
    }

    LOCK_MOZJS;

    ourl = pjs_url_to_jsobj(cx, urlp, condp);
    param[0] = OBJECT_TO_JSVAL(ourl);
    param[1] = INT_TO_JSVAL(condp->level);

    rv = JSVAL_FALSE;

    if (!JS_CallFunctionName(cx, go, "pavuk_url_cond_check", 2, param, &rv))
    {
        ret = -1;
    }
    else
    {
        /* be rather tolerant on return value; at least support non-NIL(!) integers next to the desirable boolean. */
        int32 iv = -1;
        JSBool r = JS_ValueToECMAInt32(cx, rv, &iv);
        ret = (iv < 0 || r != JS_FALSE ? -1 : iv);
    }

    JS_GC(cx);

    UNLOCK_MOZJS;

    return ret;
}


/*
 * Execute a -fnrules 'jsf' filter expression by running the JavaScript
 * function <name>(obj) where 'obj' is a PavukFnrules object.
 *
 * Returns the string produced by the JavaScript function <name> on success
 * or NULL on error.
 */
char *pjs_run_fnrules_func(char *name, struct lfname_lsp_interp *interp)
{
    char *ret = NULL;
    JSObject *rl;
    jsval param[1];
    JSString *s;
    jsval rv;

    DEBUG_JS("pjs_run_fnrules_func(%s, ...)\n", name);

    LOCK_MOZJS;
    if (pjs_func_exist(cx, go, name))
    {
        rl = pjs_fnrules_to_jsobj(cx, interp);
        param[0] = OBJECT_TO_JSVAL(rl);

        rv = JSVAL_FALSE;

        if (JS_CallFunctionName(cx, go, name, 1, param, &rv))
        {
            s = JS_ValueToString(cx, rv);
            if (s)
            {
                ret = tl_strdup(JS_GetStringBytes(s));
            }
        }

        JS_GC(cx);
    }
    else
    {
        xprintf(1, gettext("pjs: There is no \"%s\" JavaScript function!\n"), name);
    }
    UNLOCK_MOZJS;

    return ret;
}


/*
 * Execute the JavaScript function 'pavuk_parse_content(content, type, mimetype, URL, level)'
 * for the given context of type 'type' (html/css/JavaScript/other)
 * and fetched from URL 'url' at 'level'.
 *
 * mimetype, URL and level may be optional...
 *
 * Return -1 when an unsupported situation is found, for example
 * when the JavaScript function does not exist in the provided
 * JavaScript file (loaded elsewhere using '-js_script_file')
 *
 * Otherwise return an boolean/integer value as returned by the
 * JavaScript function: 0 means the content is unchanged.
 *
 * Script can call global function 'add_url_to_queue(url)' to add
 * decoded URLs to the queue. 'url' can be specified in both the
 * extended '-request' format and as a regular URL.
 */
int pjs_run_parse_content_func(char *content, size_t content_length, char *type_id, char *mime_type, url *urlp, int level)
{
    jsval param[5];
    jsval rv;
    JSObject *ourl;
    int ret = -1;

#if defined(HAVE_DEBUG_FEATURES)
    if (cfg.debug && (DEBUG_PROTOD_BITMASK & cfg.debug_level))
    {
        char dumpbuf[TEXTDUMP_BUFSIZE];

        DEBUG_JS("pjs_run_parse_content_func(<content>, %ld, %s, %s, ...) where <content>: (len = %ld) ********\n"
                 "%s\n"
                 "***************************************************\n",
                 (long)content_length, type_id, mime_type,
                 (long)content_length, tl_asciidump_content(dumpbuf, NUM_ELEM(dumpbuf), content, content_length));
    }
#endif

    if (!pjs_have_pavuk_parse_content_check || !script)
    {
        return -1;
    }

    LOCK_MOZJS;

    ourl = pjs_url_to_jsobj(cx, urlp, NULL);
    param[0] = STRING_TO_JSVAL(content);
    param[1] = STRING_TO_JSVAL(type_id);
    param[2] = STRING_TO_JSVAL(mime_type);
    param[3] = OBJECT_TO_JSVAL(ourl);
    param[4] = INT_TO_JSVAL(level);

    rv = JSVAL_FALSE;

    if (!JS_CallFunctionName(cx, go, "pavuk_parse_content", 5, param, &rv))
    {
        ret = -1;
    }
    else
    {
        /* be rather tolerant on return value; at least support non-NIL(!) integers next to the desirable boolean. */
        int32 iv = -1;
        JSBool r = JS_ValueToECMAInt32(cx, rv, &iv);
        ret = (iv < 0 || r != JS_FALSE ? -1 : iv);
    }

    JS_GC(cx);

    UNLOCK_MOZJS;

    return ret;
}


const char *pjs_GetJaveScriptLibraryVersion(void)
{
    return JS_GetImplementationVersion();
}


int pjs_run_parse_css_content_func(html_parser_t *hpinfo, html_extract_info_t *einfo)
{
    DEBUG_JS("pjs_run_parse_css_content_func(%s)\n", hpinfo->stack);

    LOCK_MOZJS;

    UNLOCK_MOZJS;

    return 0;
}

int pjs_run_parse_raw_html_content_func(html_parser_t *hpinfo, html_extract_info_t *einfo)
{
    /*
     * WARNING: hpinfo->stack has not been set for this one!
     */

    ASSERT(hpinfo->in_content ? hpinfo->in_content[hpinfo->in_size] == 0 : TRUE);
#if defined(HAVE_DEBUG_FEATURES)
    if (cfg.debug && (DEBUG_PROTOD_BITMASK & cfg.debug_level))
    {
        char dumpbuf[TEXTDUMP_BUFSIZE];

        DEBUG_JS("pjs_run_parse_raw_html_content_func (len = %ld) ********\n"
                 "%s\n"
                 "***************************************************\n",
                 (long)hpinfo->in_size, tl_asciidump_content(dumpbuf, NUM_ELEM(dumpbuf), hpinfo->in_content, hpinfo->in_size));
    }
#endif

    LOCK_MOZJS;

    UNLOCK_MOZJS;

    return 0;
}

int pjs_run_parse_js_content_func(html_parser_t *hpinfo, html_extract_info_t *einfo)
{
    DEBUG_JS("pjs_run_parse_js_content_func(%s)\n", hpinfo->stack);

    LOCK_MOZJS;

    UNLOCK_MOZJS;

    return 0;
}


/*
 * Execute the JavaScript function 'pavuk_document_process(document, phase)'
 * for the given document/URL.
 *
 * Return -1 when an unsupported situation is found, for example
 * when the JavaScript function does not exist in the provided
 * JavaScript file <cfg.js_script_file>
 *
 * Otherwise return an boolean/integer value as returned by the
 * JavaScript function.
 */
int pjs_run_doc_process_func(doc *docp, pjs_phase_t phase)
{
    jsval param[2];
    jsval rv;
    JSObject *ourl;
    int ret = -1;

    DEBUG_JS("pjs_run_doc_process_func(...)\n");

#if 0
    if (!pjs_have_pavuk_doc_process_func_check || !script)
    {
        return -1;
    }

    LOCK_MOZJS;

    ourl = pjs_url_to_jsobj(cx, urlp, condp);
    param[0] = OBJECT_TO_JSVAL(ourl);
    param[1] = INT_TO_JSVAL(condp->level);

    rv = JSVAL_FALSE;

    if (!JS_CallFunctionName(cx, go, "pavuk_document_process", 2, param, &rv))
    {
        ret = -1;
    }
    else
    {
        /* be rather tolerant on return value; at least support non-NIL(!) integers next to the desirable boolean. */
        int32 iv = -1;
        JSBool r = JS_ValueToECMAInt32(cx, rv, &iv);
        ret = (iv < 0 || r != JS_FALSE ? -1 : iv);
    }

    JS_GC(cx);

    UNLOCK_MOZJS;
#endif

    return ret;
}




#endif /* HAVE_MOZJS */


/*
 * add urls from script directly.
 *
 * use stuff like this:
 *
 * void cat_links_to_url_list(dllist * l1)
 *
 *
 * to parse the URLs in there, do not use url_parse() but the more flexible:
 *
 *
 *         // heuristics: see if we can decode the line in '-request' style,
 *         // if not, assume it is an old-style url with no frills.
 *
 *        url_info *ui = url_info_parse(mopt.args[0],
 *                                      TRUE // do not yak when this goes wrong
 *                                                                              );
 *        if (!ui)
 *        {
 *          ui = url_info_new(mopt.args[0]);
 *        }
 *            ASSERT(ui != NULL); // url_info_new() will always cough up a 'ui' struct. If the URL is legal is something entirely different.
 *
 *                and convert that to an 'url' type after that.
 *
 */

