/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#ifndef _ntlm_auth_h_
#define _ntlm_auth_h_


#ifdef ENABLE_NTLM

/****************************************************************/
/* based on document "NTLM Athentication Scheme for HTTP"       */
/* http://www.innovation.ch/java/ntlm.html                      */
/* by Ronald Tschaler <ronald@innovation.ch>                    */
/****************************************************************/

#ifdef __GNUC__
#define _PACKED         __attribute__((packed))
#else
#define _PACKED
#endif

#define NTLM_MSG1_SIZE (32 + 128)
#define NTLM_MSG2_SIZE (40)
#define NTLM_MSG3_SIZE (64 + 512)

#include "pvk_pshpack1.h"
typedef struct
{
    uint8_t protocol[8];        /* "NTLMSSP\0"                          */
    uint8_t type;               /* 0x01                                 */
    uint8_t zero[3];            /* \0\0\0                               */
    uint16_t flags;             /* 0xb203                               */
    uint8_t zero2[2];           /* \0\0                                 */

    int16_t dom_len;            /* length of domainname                 */
    int16_t dom_len2;           /* length of domainname                 */
    int16_t dom_off;            /* offset of domainname in structure    */
    uint8_t zero3[2];           /* \0\0                                 */

    int16_t host_len;           /* length of hostname                   */
    int16_t host_len2;          /* length of hostname                   */
    int16_t host_off;           /* offset of hostname in structure      */
    uint8_t zero4[2];           /* \0\0                                 */

    uint8_t padding[128];       /* space for hostname and domainname    */
#if 0
    uint8_t host[*];            /* hostname                             */
    uint8_t dom[*];             /* domainname                           */
#endif
} _PACKED ntlm_type_1_msg_t;
#include "pvk_poppack.h"

#include "pvk_pshpack1.h"
typedef struct
{
    uint8_t protocol[8];        /* "NTLMSSP\0"                          */
    uint8_t type;               /* 0x02                                 */
    uint8_t zero[7];            /* \0\0\0\0\0\0\0                       */
    int16_t msg_len;            /* 0x28                                 */
    uint8_t zero2[2];           /* \0\0                                 */
    uint16_t flags;             /* 0x8201                               */
    uint8_t zero3[2];           /* \0\0                                 */
    uint8_t nonce[8];           /* server nonce                         */
    uint8_t zero4[8];           /* \0\0\0\0\0\0\0\0                     */
} _PACKED ntlm_type_2_msg_t;
#include "pvk_poppack.h"

#include "pvk_pshpack1.h"
typedef struct
{
    uint8_t protocol[8];        /* "NTLMSSP\0"                          */
    uint8_t type;               /* 0x03                                 */
    uint8_t zero[3];            /* \0\0\0                               */

    int16_t lm_resp_len;        /* LM response length (0x18)            */
    int16_t lm_resp_len2;       /* LM response length (0x18)            */
    int16_t lm_resp_off;        /* offset of LM response in structure   */
    uint8_t zero2[2];           /* \0\0                                 */

    int16_t nt_resp_len;        /* NT response length (0x18)            */
    int16_t nt_resp_len2;       /* NT response length (0x18)            */
    int16_t nt_resp_off;        /* offset of NT response in structure   */
    uint8_t zero3[2];           /* \0\0                                 */

    int16_t dom_len;            /* length of domainname                 */
    int16_t dom_len2;           /* length of domainname                 */
    int16_t dom_off;            /* offset of domainname in structure    */
    uint8_t zero4[2];           /* \0\0                                 */

    int16_t user_len;           /* length of username                   */
    int16_t user_len2;          /* length of username                   */
    int16_t user_off;           /* offset of username in structure      */
    uint8_t zero5[2];           /* \0\0                                 */

    int16_t host_len;           /* length of hostname                   */
    int16_t host_len2;          /* length of hostname                   */
    int16_t host_off;           /* offset of hostname in structure      */
    uint8_t zero6[6];           /* \0\0\0\0\0\0                         */

    int16_t msg_len;            /* message length                       */
    uint8_t zero7[2];           /* \0\0                                 */

    uint16_t flags;             /* 0x8201                               */
    uint8_t zero8[2];           /* \0\0                                 */

    uint8_t padding[512];       /* space for hostname and domainname    */
#if 0
    uint8_t host[*];            /* hostname                             */
    uint8_t user[*];            /* username                             */
    uint8_t dom[*];             /* domainname                           */
    uint8_t lm_resp[*];         /* LM response                          */
    uint8_t nt_resp[*];         /* NT response                          */
#endif
} _PACKED ntlm_type_3_msg_t;
#include "pvk_poppack.h"

extern char *ntlm_get_t1msg_str(const char *domain, const char *host);
extern char *ntlm_get_nonce(const char *t2msg, uint16_t *t2flags);
extern char *ntlm_get_t3msg_str(const char *domain, const char *host, const char *user, const char *pass, const char *nonce, uint16_t t2flags);

extern int ntlm_negotiate_connection(doc *docp, const char *authtag);
extern int ntlm_negotiate_proxy_connection(doc *docp, const char *authtag);


#endif /* ENABLE_NTLM */
#endif

