/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "url.h"
#include "doc.h"
#include "tools.h"
#include "html.h"
#include "http.h"
#include "ftp.h"
#include "myssl.h"
#include "abstract.h"
#include "recurse.h"
#include "mime.h"
#include "robots.h"
#include "mode.h"
#include "times.h"
#include "stats.h"
#include "errcode.h"
#include "cookie.h"
#include "log.h"
#include "gui_api.h"
#include "form.h"
#include "ainterface.h"
#include "gcinfo.h"
#include "jsbind.h"




#define SETNEXTURL                \
    doc_cleanup(docu);              \
    _free(pstr);                    \
    DEBUG_PROCE(NULL);              \
    return docu->errcode



static void reschedule_url(url *urlp)
{
    DEBUG_PROCS(NULL);

    DEBUG_MISC(gettext("Rescheduling locked URL as no. %ld\n"), cfg.total_cnt);

    LOCK_CFG_URLSTACK;
    cfg.urlstack = dllist_append(cfg.urlstack, (dllist_t)urlp);
#ifdef HAVE_MT
    DEBUG_MTTHR(DBGvars("mt_semaphore_up(cfg.urlstack_sem: v = %ld BEFORE)\n"), cfg.urlstack_sem.v);
    mt_semaphore_up(&cfg.urlstack_sem);
#endif
    UNLOCK_CFG_URLSTACK;
    LOCK_TCNT;
    cfg.total_cnt++;
    UNLOCK_TCNT;

    DEBUG_PROCE(NULL);
}

static void run_post_command(doc *docp)
{
    char *urlstr;
    char *cmd = NULL;

    DEBUG_PROCS(NULL);

    DEBUG_MISC(gettext("Running post-processing command\n"));
    urlstr = url_to_urlstr(docp->doc_url, TRUE);

    cmd = tl_str_concat(DBGvars(NULL), priv_cfg.post_cmd, " \"",
                      url_to_filename(docp->doc_url, FALSE),
                      docp->is_parsable ? "\" 1 \"" : "\" 0 \"", urlstr, "\"", NULL);

    _free(urlstr);

    tl_system(cmd);

    _free(cmd);

    DEBUG_PROCE(NULL);
}

/*
 * [i_a] 'ui' may be NULL: will fill all fields from HTML form only
 */
static void add_matching_form(doc *docp, int nform, url_info *ui)
{
    char *ftext;
    int flen;
    form_info *fi;
    dllist *ptr, *fields, *sfields;
    cookie_entry *cookies = NULL;
    url_info *nui;

    DEBUG_PROCS(NULL);

    if (!(ftext = form_get_text(nform, docp->contents, docp->size, &flen)))
    {
        DEBUG_PROCE(NULL);
        return;
    }

    fi = form_parse(ftext, flen, ui);
    if (!fi)
    {
        DEBUG_PROCE(NULL);
        return;
    }

    /* copy all fields supplied on cmdln */
    fields = NULL;
    cookies = NULL;
    if (ui)
    {
        cookie_entry *cptr;

        for (ptr = ui->fields; ptr; ptr = ptr->next)
        {
            fields = dllist_prepend(fields, (dllist_t)form_field_duplicate((form_field *)ptr->data));
        }
        for (cptr = ui->cookies; cptr; cptr = cptr->next)
        {
            cookie_entry *c = dup_cookie(cptr);
            c->next = cookies;
            if (cookies)
            {
                cookies->prev = c;
            }
            cookies = c;
        }
    }

    /* copy all suitable fields from HTML form */
    sfields = NULL;
    form_get_default_successful(NULL, fi->infos, &sfields);

    for (; sfields; sfields = dllist_remove_entry(sfields, sfields))
    {
        form_field *ff = (form_field *)sfields->data;

        if (dllist_find2(fields, (dllist_t)ff, form_field_compare_name))
        {
            _free(ff->name);
            _free(ff->value);
            _free(ff);
        }
        else
        {
            fields = dllist_prepend(fields, (dllist_t)ff);
        }
    }

    nui = url_info_new(fi->action);
    nui->type = URLI_FORM;
    nui->fields = fields;
    nui->cookies = cookies;
    nui->encoding = fi->encoding;
    nui->method = fi->method;
    if (ui)
    {
        nui->localname = tl_strdup(ui->localname);
    }
    else
    {
        nui->localname = NULL;
    }

    form_free(fi);

#ifdef HAVE_DEBUG_FEATURES
    if (cfg.debug && (DEBUG_HTMLFORM_BITMASK & cfg.debug_level))
    {
        char *dbgmsg = cvt_url_info2diag_str(nui);
        DEBUG_HTMLFORM("add_matching_form[%d]: before append_starting_url: URL INFO = %s", nform, dbgmsg);
        _free(dbgmsg);
    }
#endif
    VERIFY(append_starting_url(nui, docp->doc_url));
#ifdef HAVE_DEBUG_FEATURES
    if (cfg.debug && (DEBUG_HTMLFORM_BITMASK & cfg.debug_level))
    {
        char *dbgmsg = cvt_url_info2diag_str(nui);
        DEBUG_HTMLFORM("add_matching_form[%d]: after append_starting_url: URL INFO = %s", nform, dbgmsg);
        _free(dbgmsg);
    }
#endif

    url_info_free(nui);
    nui = NULL;

    DEBUG_PROCE(NULL);
}

static void add_matching_forms(doc *docp, dllist *formlist)
{
    dllist *fptr, *uptr;
    int nform;

    DEBUG_PROCS(NULL);

    for (fptr = formlist, nform = 0; fptr; fptr = fptr->next, nform++)
    {
        url *urlp;
        bool_t found = FALSE;

        DEBUG_HTMLFORM("add_matching_forms[%d]: '%s'\n", nform, (char *)fptr->data);

        urlp = url_parse((char *)fptr->data);

        ASSERT(urlp->type != URLT_FROMPARENT);

        if ((urlp->type != URLT_HTTP) && (urlp->type != URLT_HTTPS))
        {
            free_deep_url(urlp);
            _free(urlp);
            continue;
        }
        free_deep_url(urlp);
        _free(urlp);

        for (uptr = priv_cfg.formdata; uptr; uptr = uptr->next)
        {
            url_info *ui = (url_info *)uptr->data;

            DEBUG_HTMLFORM("forms match check: '%s' =?= '%s'\n", ui->urlstr, (char *)fptr->data);

            if (!strcmp(ui->urlstr, (char *)fptr->data))
            {
                DEBUG_HTMLFORM("forms match check: Added matched form: '%s'\n", ui->urlstr);
                /* include form multiple times, if matched by different formdata items */
#ifdef HAVE_DEBUG_FEATURES
                if (cfg.debug && (DEBUG_HTMLFORM_BITMASK & cfg.debug_level))
                {
                    char *dbgmsg = cvt_url_info2diag_str(ui);
          DEBUG_HTMLFORM("add_matching_forms[%d]: Added matched form: URL INFO = %s",
                         nform,
                         dbgmsg);
                    _free(dbgmsg);
                }
#endif
                add_matching_form(docp, nform, ui);
                found = TRUE;
            }
        }

        /* [i_a] addition: just add all non-matched forms too! */
        if (!found)
        {
      DEBUG_HTMLFORM("%s unmatched form '%s'\n", (cfg.try_all_forms ? "Added" : "Skipped"),
                     (char *)fptr->data);
            if (cfg.try_all_forms)
            {
                add_matching_form(docp, nform, NULL);
            }
        }
    }

    DEBUG_PROCE(NULL);
}

int process_document(doc *docu, int check_lim)
{
    url *urlr;
    int nreget = 0;
    int nredir = 0;
    int pokus = 0;
    time_t atm;
    char cpom[64];
    char *pstr = NULL;
    struct stat estat;

    DEBUG_PROCS(docu && docu->doc_url && docu->doc_url->urlstr ? docu->doc_url->urlstr : "???");

    urlr = docu->doc_url;

    docu->check_limits = check_lim;

    gui_loop_serve();

    if (cfg.debug && (DEBUG_LIMITS_BITMASK & cfg.debug_level))
    {
        char *urldbgstr = NULL;

        cvt_url2diag_str(urlr, &urldbgstr, NULL);
        DEBUG_LIMITS("process_document(): check against limits? "
                     "check: %s + has.parent: %s --> check: %s for:\n"
                     "%s\n"
                     "-----------\n",
                     (docu->check_limits ? "YES" : "NO"), ((urlr && urlr->parent_url) ? "YES" : "NO"),
      ((docu->check_limits && urlr && urlr->parent_url) ? "YES" : "NO"),
urldbgstr);
        _free(urldbgstr);
    }

    if (docu->check_limits)
    {
        docu->check_limits = (urlr->parent_url != NULL);
    }

    while (!cfg.stop && !cfg.rbreak)
    {
        _free(docu->ftp_pasv_host);
        docu->errcode = ERR_NOERROR;
        docu->mime = NULL;
        docu->type_str = NULL;
        docu->doc_url = urlr;
        docu->dtime = 0L;
        docu->contents = NULL;
        docu->request_header = NULL;
        docu->request_data = NULL;
        docu->request_datalen = 0;
//    docu->actrec_ref = NULL;
        docu->is_chunked = FALSE;
        docu->read_chunksize = FALSE;
        docu->read_trailer = FALSE;
        docu->ftp_fatal_err = FALSE;
        pstr = url_to_urlstr(urlr, FALSE);

        if (pokus)
        {
            xprintf(1, gettext("retry no. %d/%ld\n"), pokus, cfg.nretry);
        }

        if (cfg.progress || cfg.verbose)
        {
            char *msg = NULL;

            if (cfg.progress_mode < 10)
            {
#ifdef HAVE_MT
        msg = tl_str_printf(NULL, gettext("URL[%2d]: %5d(%ld) of %5ld  %s\n"), docu->threadnr + 1, docu->doc_nr, cfg.fail_cnt, cfg.total_cnt, pstr);
#else
                msg = tl_str_printf(NULL, gettext("URL: %5d(%ld) of %5ld  %s\n"), docu->doc_nr, cfg.fail_cnt, cfg.total_cnt, pstr);
#endif
            }
            print_progress_style(1, cfg.progress_mode, msg,
#ifdef HAVE_MT
                                 docu->threadnr + 1,
#else
                                 0,
#endif
                                 docu->is_persistent);
            _free(msg);
        }

#ifdef I_FACE
        if (cfg.xi_face)
        {
            gui_set_doccounter();

            gui_set_url(pstr);

            gui_set_status(gettext("Starting download"));
        }
#endif

        /*** to be able to revisit moved documents ***/
        /*** especially for authorization purposes ***/
        if ((urlr->status & URL_PROCESSED) && urlr->moved_to && nredir)
        {
            urlr->status &= ~URL_PROCESSED;
        }

DEBUG_TRACE("process_document(): check against limits level CL2? %s\n", 
      (docu->check_limits ? "YES" : "NO"));

        if (docu->check_limits)
        {
            cond_info_t condp = { 0 };

            /*
             * level == 2 (CL2): condition checking is called when URL is taken
             * from download queue and will be transfered after this
             * check will be successful.
             *
             * Only URLs which were *collected* during the spider operation will
             * have this level of limit checks; any URLs which were specified
             * at the command line or in the GUI will *NOT* have these checks
             * performed as their 'check_limits' flag == FALSE!
             */
            condp.level = 2;
            condp.urlnr = docu->doc_nr;
            condp.size = 0;
            condp.time = 0L;
            condp.mimet = NULL;
            condp.full_tag = NULL;
            condp.params = NULL;
            condp.html_doc = NULL;
            condp.html_doc_offset = 0;
            condp.tag = NULL;
            condp.attrib = NULL;

            if (urlr->status & URL_PROCESSED)
            {
                xprintf(1, gettext("Already processed\n"));
                docu->errcode = ERR_PROCESSED;
                SETNEXTURL;
            }

            if (urlr->status & URL_USER_DISABLED)
            {
                xprintf(1, gettext("Disallowed by user\n"));
                docu->errcode = ERR_UDISABLED;
                SETNEXTURL;
            }

            ASSERT(urlr->type >= 0);
            ASSERT(urlr->type < prottable_size());
            if (!prottable[urlr->type].supported
               || (urlr->parent_url
                  && (urlr->type == URLT_FTP
                     || urlr->type == URLT_FTPS)
              && urlr->p.ftp.dir
              && !cfg.condition.ftpdir)
          || (urlr->parent_url
              && !url_append_condition(urlr, &condp)))
            {
                if (!cfg.verbose)
                {
                    xprintf(1, gettext("Disallowed by rules\n"));
                }
                else
                {
                    char *urlstr = NULL;
                    char *urldbgstr = NULL;
                    char *typ_str = "UNKNOWN";

                    switch (urlr->type)
                    {
                    case URLT_HTTP:
                        typ_str = "HTTP";
                        break;

                    case URLT_HTTPS:
                        typ_str = "HTTPS";
                        break;

                    case URLT_FTP:
                        typ_str = "FTP";
                        break;

                    case URLT_FTPS:
                        typ_str = "FTPS";
                        break;

                    case URLT_FILE:
                        typ_str = "FILE";
                        break;

                    case URLT_GOPHER:
                        typ_str = "GOPHER";
                        break;

                    case URLT_FROMPARENT:
                        typ_str = "FROMPARENT";
                        break;

                    default:
                        ASSERT(strcmp(typ_str, "UNKNOWN") == 0);
                        break;
                    }

                    cvt_url2diag_str(urlr, &urldbgstr, &urlstr);
          xprintf(1, gettext("Disallowed by rules:\n%s -- Supported type: %d (%s)\n"), 
			  urldbgstr, prottable[urlr->type].supported, typ_str);
                    _free(urldbgstr);
                    _free(urlstr);
                }

                urlr->status |= URL_REJECTED;
                docu->errcode = ERR_RDISABLED;
                SETNEXTURL;
            }

            if (!robots_check(urlr))
            {
                urlr->status |= URL_REJECTED;
                docu->errcode = ERR_RDISABLED;
                SETNEXTURL;
            }
        }

        if (cfg.mode == MODE_FTPDIR && (urlr->type != URLT_FTP && urlr->type != URLT_FTPS))
        {
            xprintf(1, gettext("This URL type is not supported with ftpdir mode\n"));

            urlr->status |= URL_REJECTED;
            docu->errcode = ERR_RDISABLED;
            SETNEXTURL;
        }

        gui_loop_serve();

        if (cfg.mode == MODE_SYNC)
        {
            const char *pp = url_to_filename(urlr, TRUE);

            if (!tl_stat(pp, &estat) && !S_ISDIR(estat.st_mode))
            {
                atm = time(NULL) - 86400 * cfg.ddays;
                /*
                 * pro: We do not want the message
                 * "No transfer - file not expired"
                 * if the server's clock is ahead of our clock.
                 * If no parameter cfg.ddays is given, then
                 * we do not compare the file modification times.
                 */
                if (cfg.ddays == 0 || estat.st_mtime < atm)
                {
                    docu->dtime = estat.st_mtime;
                }
                else
                {
                    xprintf(1, gettext("No transfer - file not expired\n"));
                    urlr->status |= URL_REJECTED;
                    docu->errcode = ERR_RDISABLED;
                    SETNEXTURL;
                }
                urlr->status |= URL_ISLOCAL;
                docu->origsize = estat.st_size;
            }
        }

        if (cfg.show_time)
        {
            atm = time(NULL);
            LOCK_TIME;
            strftime(cpom, sizeof(cpom), "%H:%M:%S", localtime(&atm));
            UNLOCK_TIME;
            xprintf(1, gettext("Starting time :  %s\n"), cpom);
        }

#ifdef I_FACE
        if (cfg.stop || cfg.rbreak)
        {
            _free(pstr);
            break;
        }
#endif

        gui_loop_serve();

        if ((urlr->type == URLT_FTP || urlr->type == URLT_FTP)
           && urlr->extension
        && ((ftp_url_extension *)urlr->extension)->type == FTP_TYPE_L
        && ((ftp_url_extension *)urlr->extension)->slink)
        {
            if (cfg.retrieve_slink)
            {
                /** need to kill extension, because we must **/
                /** guess the file type beside the symlink  **/
                ftp_url_ext_free(urlr->extension);
                urlr->extension = NULL;
            }
            else
            {
                ftp_make_symlink(urlr);
                urlr->status |= URL_PROCESSED;
                docu->errcode = ERR_NOERROR;
                SETNEXTURL;
            }
        }

        gui_set_status(gettext("Starting download"));

        ASSERT(!docu->contents);
        if (doc_download(docu, FALSE, FALSE))
        {
#if defined(HAVE_DEBUG_FEATURES)
            ASSERT(docu->contents ? docu->contents[docu->size] == 0 : TRUE);
            if (cfg.debug && (DEBUG_PROTOD_BITMASK & cfg.debug_level))
            {
                char dumpbuf[TEXTDUMP_BUFSIZE];

                DEBUG_PROTOD("********************* Response: Content (len = %ld) ********\n"
                             "%s\n"
                             "***************************************************\n",
                             (long)docu->size, tl_asciidump_content(dumpbuf, NUM_ELEM(dumpbuf), docu->contents, docu->size));
            }
#endif

            if (cfg.show_time)
            {
                atm = time(NULL);
                LOCK_TIME;
                strftime(cpom, sizeof(cpom), "%H:%M:%S", localtime(&atm));
                UNLOCK_TIME;
                xprintf(1, gettext("Ending time :    %s\n"), cpom);
            }

            gui_loop_serve();

            doc_remove_lock(docu);
            _free(docu->contents);
            _free(docu->request_header);
            _free(docu->request_data);
            //     docu->actrec_ref = NULL;

            DEBUG_USER("Error status code - (%d)\n", docu->errcode);
            report_error(docu, gettext("download"));

            if ((nreget < cfg.nreget
                && (docu->errcode == ERR_HTTP_TRUNC
                   || docu->errcode == ERR_FTP_TRUNC
                   || docu->errcode == ERR_LOW_TRANSFER_RATE
                   || docu->errcode == ERR_HTTP_FAILREGET
                   || docu->errcode == ERR_HTTP_TIMEOUT
                   || docu->errcode == ERR_HTTP_GW_TIMEOUT))
               || (nredir < cfg.nredir
              && docu->errcode == ERR_HTTP_REDIR)
          || (docu->errcode == ERR_HTTP_AUTH)
          || (docu->errcode == ERR_HTTP_PROXY_AUTH))
            {
                if (docu->errcode == ERR_HTTP_REDIR)
                {
                    urlr->status |= URL_PROCESSED;
                    if ((urlr->moved_to->status & URL_PROCESSED) && (!urlr->moved_to->moved_to))
                    {
                        SETNEXTURL;
                    }
                    else
                    {
#ifdef I_FACE
                        if (cfg.xi_face)
                        {
                            gui_tree_set_icon_for_doc(docu);
                        }
#endif
                        urlr = urlr->moved_to;
                    }
                }

                if (docu->errcode == ERR_HTTP_TRUNC)
                {
                    urlr->status |= URL_TRUNCATED;
                    _free(docu->etag);

                    docu->etag = get_mime_param_val_str("ETag:", docu->mime);
                    if (!docu->etag)
                    {
                        docu->etag = get_mime_param_val_str("Content-Location:", docu->mime);
                    }
                    if (!docu->etag)
                    {
                        docu->etag = get_mime_param_val_str("Last-Modified", docu->mime);
                    }
                }

                if (docu->errcode == ERR_HTTP_AUTH)
                {
                    docu->doc_url->status |= URL_PROCESSED;
                    docu->doc_url->status |= URL_ERR_REC;
                    SETNEXTURL;
                }

                if (docu->errcode == ERR_HTTP_PROXY_AUTH)
                {
                    docu->doc_url->status |= URL_PROCESSED;
                    docu->doc_url->status |= URL_ERR_REC;
                    SETNEXTURL;
                }

                _free(docu->mime);
                _free(docu->type_str);

                /* [i_a] also count a total failure to progress a reget operation as 'one round'
                 * to prevent infinite loops due to a server found incapable of handling a
                 * REST/reget operation. */
        nreget += (docu->errcode == ERR_HTTP_TRUNC ||
                   docu->errcode == ERR_HTTP_FAILREGET ||
                   docu->errcode == ERR_FTP_TRUNC) && cfg.mode != MODE_SREGET;
                nredir += (docu->errcode == ERR_HTTP_REDIR);
                _free(pstr);
                continue;
            }

            if (docu->errcode == ERR_FTP_UNKNOWN
               || docu->errcode == ERR_FTP_CONNECT
               || docu->errcode == ERR_FTP_DATACON
               || docu->errcode == ERR_FTPS_CONNECT
               || docu->errcode == ERR_FTPS_DATASSLCONNECT
               || docu->errcode == ERR_HTTP_UNKNOWN
               || docu->errcode == ERR_HTTP_CONNECT
               || docu->errcode == ERR_HTTP_SNDREQ
               || docu->errcode == ERR_HTTP_SNDREQDATA
               || docu->errcode == ERR_HTTP_RCVRESP
               || docu->errcode == ERR_HTTP_SERV
               || docu->errcode == ERR_HTTP_TIMEOUT
               || docu->errcode == ERR_HTTP_PROXY_CONN
               || docu->errcode == ERR_HTTPS_CONNECT
               || docu->errcode == ERR_READ
               || docu->errcode == ERR_ZERO_SIZE
          || docu->errcode == ERR_GOPHER_CONNECT
          || docu->errcode == ERR_PROXY_CONNECT
          || docu->errcode == ERR_HTTP_SERV)
            {
                urlr->status |= URL_ERR_REC;
                pokus++;
                /*** retry only when allowed ***/
                if (pokus >= cfg.nretry)
                {
                    urlr->status |= URL_PROCESSED;
                    SETNEXTURL;
                }
                _free(pstr);
                _free(docu->mime);
                _free(docu->type_str);
                continue;
            }
            else if (docu->errcode == ERR_LOCKED)
            {
                LOCK_CFG_URLSTACK;
                if (!cfg.urlstack)
                {
                    UNLOCK_CFG_URLSTACK;
                    xprintf(1, gettext("last document locked -> sleeping for 5 seconds\n"));
                    tl_sleep(5);
                }
                else
                {
                    UNLOCK_CFG_URLSTACK;
                }
                reschedule_url(urlr);
                SETNEXTURL;
            }
            else if (docu->errcode == ERR_BIGGER
                    || docu->errcode == ERR_SMALLER
               || docu->errcode == ERR_NOMIMET
               || docu->errcode == ERR_OUTTIME 
               || docu->errcode == ERR_SCRIPT_DISABLED)
            {
                urlr->status |= URL_PROCESSED;
                urlr->status |= URL_ERR_REC;
                SETNEXTURL;
            }
            else
            {
                /*** remove improper documents if required ***/
                if ((cfg.remove_old
                    && (cfg.mode == MODE_SYNC
                       || cfg.mode == MODE_MIRROR))
                   && (((docu->errcode == ERR_FTP_GET
                        || docu->errcode == ERR_FTP_BDIR
                        || docu->errcode == ERR_FTP_NODIR)
                 && docu->ftp_respc == FTP_RSP_ERR550)
                || docu->errcode == ERR_HTTP_NFOUND
                || docu->errcode == ERR_HTTP_GONE))
                {
                    doc_remove(docu->doc_url);
                }

                urlr->status |= URL_ERR_UNREC;
                urlr->status |= URL_PROCESSED;

                SETNEXTURL;
            }
        }
        else
        {
#ifdef HAVE_DEBUG_FEATURES
            ASSERT(docu->contents ? docu->contents[docu->size] == 0 : TRUE);
            if (cfg.debug && (DEBUG_PROTOD_BITMASK & cfg.debug_level))
            {
                char dumpbuf[TEXTDUMP_BUFSIZE];

                DEBUG_PROTOD("************** Response: Content OK (len = %ld) ********\n"
                             "%s\n"
                             "***************************************************\n",
                             (long)docu->size, tl_asciidump_content(dumpbuf, NUM_ELEM(dumpbuf), docu->contents, docu->size));
            }
#endif
        }

        gui_loop_serve();

        if (urlr->status & URL_TRUNCATED)
        {
            urlr->status &= ~URL_TRUNCATED;
        }

        if (urlr->status & URL_ERR_REC)
        {
            urlr->status &= ~URL_ERR_REC;
        }

        if (cfg.show_time)
        {
            atm = time(NULL);
            LOCK_TIME;
            strftime(cpom, sizeof(cpom), "%H:%M:%S", localtime(&atm));
            UNLOCK_TIME;
            xprintf(1, gettext("Ending time :    %s\n"), cpom);
        }

        report_error(docu, gettext("download"));

        gui_loop_serve();

        ASSERT(docu->contents ? docu->contents[docu->size] == 0 : TRUE);
        if (docu->contents)
        {
            if (docu->is_parsable)
            {
                dllist *formlist = NULL;
                dllist *urls;

                gui_set_status(gettext("Relocating and scanning HTML/CSS/JS document"));

#ifdef HAVE_DEBUG_FEATURES
                ASSERT(docu->contents ? docu->contents[docu->size] == 0 : TRUE);
                if (cfg.debug && (DEBUG_TRACE_BITMASK & cfg.debug_level))
                {
                    char dumpbuf[TEXTDUMP_BUFSIZE];

                    DEBUG_TRACE("************** scanning HTML/CSS/JS document (len = %ld) ********\n"
                                "%s\n"
                                "***************************************************\n",
                                (long)docu->size, tl_asciidump_content(dumpbuf, NUM_ELEM(dumpbuf), docu->contents, docu->size));
                }
#endif
                urls = html_process_document(docu, priv_cfg.formdata ? &formlist : NULL);

                gui_loop_serve();

                if (urls && bufio_is_open(cfg.dump_urlfd))
                {
                    url *tmp_urlr;
                    char *tmp_pstr = NULL;
                    tmp_urlr = docu->doc_url;
                    tmp_pstr = url_to_urlstr(tmp_urlr, FALSE);
                    DEBUG_TRACE("url list dump for URL '%s'\n", tmp_pstr);

                    dump_urls_list(urls, docu);
                    _free(tmp_pstr);
                }
                if (priv_cfg.formdata && bufio_is_open(cfg.dump_fd))
                {
                    url *tmp_urlr;
                    char *tmp_pstr = NULL;
                    tmp_urlr = docu->doc_url;
                    tmp_pstr = url_to_urlstr(tmp_urlr, FALSE);
                    DEBUG_TRACE("formdata dump for URL '%s'\n", tmp_pstr);

                    dump_cfg_formdata(tmp_pstr, priv_cfg.formdata);
                    _free(tmp_pstr);
                }
                if (formlist && bufio_is_open(cfg.dump_fd))
                {
                    url *tmp_urlr;
                    char *tmp_pstr = NULL;
                    tmp_urlr = docu->doc_url;
                    tmp_pstr = url_to_urlstr(tmp_urlr, FALSE);
                    DEBUG_TRACE("formlist dump for URL '%s'\n", tmp_pstr);

                    dump_formlist(tmp_pstr, formlist);
                    _free(tmp_pstr);
                }

                if ((priv_cfg.formdata || cfg.try_all_forms) && formlist)
                {
                    add_matching_forms(docu, formlist);
                    while (formlist)
                    {
                        if (formlist->data)
                        {
                            _free(formlist->data);
                        }
                        formlist = dllist_remove_entry(formlist, formlist);
                    }
                }

        if (cfg.mode != MODE_SREGET && cfg.mode != MODE_FTPDIR &&
            !(docu->doc_url->status & URL_NORECURSE))
                {
#if defined I_FACE
                    gui_tree_add_start();
#endif
                    cat_links_to_url_list(urls);
#if defined I_FACE
                    gui_tree_add_end();
#endif
                }
                else if (cfg.mode == MODE_FTPDIR)
                {
                    dump_ftp_list(urls);
                }
                else
                {
                    while (urls)
                    {
                        free_deep_url((url *)urls->data);
                        _free(urls->data);
                        urls = dllist_remove_entry(urls, urls);
                    }
                }

                gui_loop_serve();
            }

            pjs_run_doc_process_func(docu, PJS_DPH_STORE);

            if (bufio_is_open(cfg.dump_fd) && cfg.dump_after)
            {
                bufio *fd;

                gui_set_status(gettext("Dumping processed document"));
                VERIFY(!bufio_lock(cfg.dump_fd, FALSE, &fd));
                if (docu->request_header && cfg.dump_req)
                {
                    bufio_asciidump(fd, docu->request_header, strlen(docu->request_header), DUMP_SECTION_SEPSTR);
                }
                if (docu->request_data && cfg.dump_req)
                {
                    bufio_asciidump(fd, docu->request_data, docu->request_datalen, NULL);
                }
                if (docu->mime && cfg.dump_resp)
                {
                    bufio_puts(fd, DUMP_SECTION_SEPSTR);
                    bufio_puts(fd, docu->mime);
                }
                ASSERT(docu->contents ? docu->contents[docu->size] == 0 : TRUE);
                if (docu->contents && cfg.dump_resp)
                {
                    bufio_asciidump(fd, docu->contents, docu->size, NULL);
                }
                VERIFY(!bufio_release(cfg.dump_fd, &fd));
            }

            if ((docu->doc_url->type != URLT_FILE)
               && !(docu->doc_url->status & URL_REDIRECT)
               && (docu->errcode != ERR_HTTP_ACTUAL)
		&& (docu->errcode != ERR_FTP_ACTUAL) && (cfg.mode != MODE_NOSTORE)
		&& (cfg.mode != MODE_FTPDIR))
            {
                int store_stat;

                gui_set_status(gettext("Storing document"));

                store_stat = doc_store(docu, TRUE);
                if (store_stat)
                {
                    xprintf(1, gettext("Store failed\n"));
                    urlr->status &= ~URL_ERR_REC;
                }
            }

            gui_loop_serve();

            if (priv_cfg.post_cmd)
            {
                run_post_command(docu);
            }

            doc_remove_lock(docu);

            doc_update_parent_links(docu);
        }
        else
        {
            pjs_run_doc_process_func(docu, PJS_DPH_STORE);

            if (priv_cfg.post_cmd)
            {
                run_post_command(docu);
            }

            doc_remove_lock(docu);

            doc_update_parent_links(docu);
        }

        urlr->status |= URL_DOWNLOADED;
        urlr->status |= URL_PROCESSED;
        SETNEXTURL;
    }

    DEBUG_PROCE(NULL);

    return ERR_UNKNOWN;
}

#ifdef I_FACE
void *download_single_doc_as_thread(void *param)
{
    url *urlp = (url *)param;
    int rv;

    VERIFY(0 == pthread_setspecific(cfg.thrnr_key, cast_int2voidptr(-2)));

    DEBUG_PROCS(NULL);

    rv = download_single_doc(urlp);

    DEBUG_MTTHR(DBGvars("download_single_doc thread exiting\n"));
    /*  gui_set_status(gettext("Exiting ...")); */

    DEBUG_PROCE(NULL);

    pthread_exit(NULL);
    return NULL;
}

int download_single_doc(url *urlp)
{
    int rv;
    doc docu;
    global_connection_info con_info;

#if defined(HAVE_MT)
    _config_struct_priv_t privcfg;

#if defined(__OSF__) || defined(__osf__)
#define __builtin_try
#define __builtin_finally
#endif

    DEBUG_PROCS(NULL);

    privcfg_make_copy(&privcfg);
    VERIFY(0 == pthread_setspecific(cfg.privcfg_key, &privcfg));
    pthread_cleanup_push(&privcfg_free4cleanup, &privcfg);
#endif

    gui_start_download();

#ifdef HAVE_MT
    {
#if defined(HAVE_SIGEMPTYSET) && defined(HAVE_SIGADDSET) && defined(HAVE_PTHREAD_SIGMASK)
        sigset_t smask;

        sigemptyset(&smask);
        sigaddset(&smask, SIGINT);
#ifdef SIGQUIT
        sigaddset(&smask, SIGQUIT);
#endif
#ifdef SIGXFSZ
        sigaddset(&smask, SIGXFSZ);
#endif
        pthread_sigmask(SIG_UNBLOCK, &smask, NULL);
#endif

        tl_signal(SIGINT, pavuk_sigintthr);
#ifdef SIGQUIT
        tl_signal(SIGQUIT, pavuk_sigquitthr);
#endif
#ifdef SIGXFSZ
        tl_signal(SIGXFSZ, pavuk_sigfilesizethr);
#endif
    }

    VERIFY(0 == pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL));
#if 0
    VERIFY(0 == pthread_setspecific(cfg.thrnr_key, cast_int2voidptr(0)));
#endif
    DEBUG_MTTHR(DBGvars("starting thread(%ld) %d\n"), pthread_t2long(pthread_self()), 0);

    cfg.allthreadsnr = 0;
    gui_mt_thread_start(cfg.allthreadsnr);
#endif

    cfg.rbreak = FALSE;
    cfg.stop = FALSE;
    cfg.processing = TRUE;

    doc_init(&docu, urlp);

#ifdef HAVE_MT
    docu.threadnr = 0;
    VERIFY(0 == pthread_setspecific(cfg.currdoc_key, NULL));
    VERIFY(0 == pthread_setspecific(cfg.herrno_key, &docu.__herrno));
#endif
#ifdef INCLUDE_CHUNKY_DoS_FEATURES
    docu.cyclenr = 0;
#endif
    rv = process_document(&docu, FALSE /* no limit checking for this guy! */);

    init_global_connection_data(&con_info);
    save_global_connection_data(&con_info, &docu);
    kill_global_connection_data(&con_info);

    cfg.processing = FALSE;

    cfg.rbreak = FALSE;
    cfg.stop = FALSE;

#ifdef HAVE_MT
    pthread_cleanup_pop(TRUE);
    doc_finish_processing(&docu);
    cfg.allthreadsnr = 0;
    gui_mt_thread_end(0);
#endif
    gui_beep();
    gui_set_msg(gettext("Done"), 0);

    DEBUG_PROCE(NULL);

    return rv;
}
#endif


/*********************************************/
/* rekurzivne prechadzanie stromu dokumentov */
/* FIXME: Translate me!                      */
/*********************************************/
#ifdef HAVE_MT
static void mt_recurse(int thrnr)
#else
void recurse(int thrnr)
#endif
{
    global_connection_info con_info;

    DEBUG_PROCS(NULL);

#if 0
    if (cfg.urlstack == NULL)
    {
        DEBUG_PROCE(NULL);
        return;
    }
#endif

    init_global_connection_data(&con_info);

    /**** obsluzenie vsetkych URL v zozname ****/
    /**** FIXME: Translate me!              ****/
    while (!cfg.stop && !cfg.rbreak)
    {
        doc docu = { 0 };
        url *urlp;

        LOCK_CFG_URLSTACK;
        /*
         * WARNING! this condition may look rediculous, given the same condition in the
         * while() above, but we're working with multithreaded code here: a different thread
         * executing this same code may nullify 'cfg.urlstack' during the time it took
         * _this_ thread to arrive here. (And believe me, that happens quite often!)
         *
         * So we check again to make sure inside this 'critical section' which is blocked off
         * by the [UN]LOCK_CFG_URLSTACK calls.
         *
         * UPDATE (aug/2008): removed the cfg.urlstack check in the while() condition
         * as it only was causing trouble due to asynchronicy when running multiple threads
         * processing urlstack entries.
         */
        if (cfg.urlstack)
        {
            urlp = (url *)cfg.urlstack->data;
            cfg.urlstack = dllist_remove_entry(cfg.urlstack, cfg.urlstack);
#ifdef HAVE_MT
            DEBUG_MTTHR(DBGvars("mt_semaphore_decrement(cfg.urlstack_sem: v = %ld BEFORE)\n"), cfg.urlstack_sem.v);
            mt_semaphore_decrement(&cfg.urlstack_sem);
#endif
            UNLOCK_CFG_URLSTACK;
            DEBUG_MTTHR(DBGvars("following UNLOCK_CFG_URLSTACK\n"));
        }
        else
        {
            UNLOCK_CFG_URLSTACK;
#ifdef HAVE_MT
            /*
             * as 'docu' is stack-based storage, make sure we destroy references pointing to that
             * storage once we unroll the stack that much. And that's right HERE!
             */
            VERIFY(0 == pthread_setspecific(cfg.currdoc_key, NULL));
            VERIFY(0 == pthread_setspecific(cfg.herrno_key, NULL));
#endif
            break;
        }

        doc_init(&docu, urlp);

#ifdef HAVE_MT
        docu.threadnr = thrnr;
        VERIFY(0 == pthread_setspecific(cfg.thrnr_key, cast_int2voidptr(thrnr)));
        VERIFY(0 == pthread_setspecific(cfg.herrno_key, &docu.__herrno));
        VERIFY(0 == pthread_setspecific(cfg.currdoc_key, &docu));
#endif

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
        docu.cyclenr = 0;
#endif

        LOCK_DCNT;
        cfg.docnr++;
        docu.doc_nr = cfg.docnr;
        UNLOCK_DCNT;

        restore_global_connection_data(&con_info, &docu);

        process_document(&docu, TRUE);

        save_global_connection_data(&con_info, &docu);

#ifdef HAVE_MT
        doc_finish_processing(&docu);
#endif

    if (docu.errcode == ERR_QUOTA_FS
        || docu.errcode == ERR_QUOTA_TRANS
        || docu.errcode == ERR_QUOTA_TIME 
		|| cfg.rbreak || cfg.stop)
        {
            LOCK_CFG_URLSTACK;
            LOCK_DCNT;
            cfg.docnr--;
            UNLOCK_DCNT;
            cfg.urlstack = dllist_prepend(cfg.urlstack, (dllist_t)urlp);
#ifdef HAVE_MT
            DEBUG_MTTHR(DBGvars("mt_semaphore_up(cfg.urlstack_sem: v = %ld BEFORE)\n"), cfg.urlstack_sem.v);
            mt_semaphore_up(&cfg.urlstack_sem);
#endif
            UNLOCK_CFG_URLSTACK;
#ifdef HAVE_MT
            /*
             * as 'fake_doc' is stack-based storage, make sure we destroy references pointing to that
             * storage once we unroll the stack that much. And that's right HERE!
             */
            VERIFY(0 == pthread_setspecific(cfg.currdoc_key, NULL));
            VERIFY(0 == pthread_setspecific(cfg.herrno_key, NULL));
#endif
            break;
        }
#ifdef HAVE_MT
        /*
         * as 'fake_doc' is stack-based storage, make sure we destroy references pointing to that
         * storage once we unroll the stack that much. And that's right HERE! (Well, not really, but
         * docu goes out of scope/comes in scope at this spot, so it's a good place to put this.
         * Putting this code /behind/ the closing curly brace would be a disaster as you'd be
         * already out of scope before you'd have a chance to patch the values to NULL. This is our
         * last chance while having a 'legal' [docu].
         */
        VERIFY(0 == pthread_setspecific(cfg.currdoc_key, NULL));
        VERIFY(0 == pthread_setspecific(cfg.herrno_key, NULL));
#endif
    }
#if defined(I_FACE) && !defined(HAVE_MT)
    if (cfg.xi_face)
    {
        gui_set_status(gettext("Done"));
    }
#endif

#ifdef I_FACE
    if (cfg.xi_face)
    {
        gui_set_doccounter();
    }
#endif

    if (cfg.update_cookies)
    {
        cookie_update_file(TRUE);
    }

    kill_global_connection_data(&con_info);

    DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "recurse() restored cfg.stop/cfg.rbreak", cfg.stop, cfg.rbreak);

    if (!cfg.rbreak && !cfg.stop && cfg.stats_file)
    {
        stats_fill_spage(cfg.stats_file, NULL);
    }

    DEBUG_PROCE(NULL);
}

#ifdef HAVE_MT
static void *mt_recurse_thrd(void *param)
{
    int thrnr = cast_voidptr2int(param);

/*  bool_t init = (thrnr == 0); */
#ifdef I_FACE
    _config_struct_priv_t privcfg;
#endif

    DEBUG_PROCS(NULL);

    {
#if defined(HAVE_SIGEMPTYSET) && defined(HAVE_SIGADDSET) && defined(HAVE_PTHREAD_SIGMASK)
        sigset_t smask;

        sigemptyset(&smask);
        sigaddset(&smask, SIGINT);
#ifdef SIGQUIT
        sigaddset(&smask, SIGQUIT);
#endif
#ifdef SIGXFSZ
        sigaddset(&smask, SIGXFSZ);
#endif
        pthread_sigmask(SIG_UNBLOCK, &smask, NULL);
#endif

        signal(SIGINT, pavuk_sigintthr);
#ifdef SIGQUIT
        signal(SIGQUIT, pavuk_sigquitthr);
#endif
#ifdef SIGXFSZ
        signal(SIGXFSZ, pavuk_sigfilesizethr);
#endif
    }

    VERIFY(0 == pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL));
#if defined(_MSC_VER)
    /*
     * warning C4312: 'type cast' : conversion from 'int' to 'void *' of greater size
     * warning C4312: 'type cast' : conversion from 'long' to 'void *' of greater size
     */
#pragma warning(push)
#pragma warning(disable: 4311 4312)
#endif
    VERIFY(0 == pthread_setspecific(cfg.currdoc_key, NULL));
    VERIFY(0 == pthread_setspecific(cfg.thrnr_key, cast_int2voidptr(thrnr)));
#if defined(_MSC_VER)
#pragma warning(pop)
#endif
    DEBUG_MTTHR(DBGvars("starting thread(%ld) %d\n"), pthread_t2long(pthread_self()), thrnr);

#ifdef I_FACE
    privcfg_make_copy(&privcfg);
    VERIFY(0 == pthread_setspecific(cfg.privcfg_key, &privcfg));
    pthread_cleanup_push(&privcfg_free4cleanup, &privcfg);
#endif

    for (; !cfg.rbreak && !cfg.stop;)
    {
        VERIFY(0 == pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL));
        DEBUG_MTTHR(DBGvars("thread %d awaking\n"), thrnr);

        mt_recurse(thrnr);
        DEBUG_MTTHR(DBGvars("thread %d sleeping\n"), thrnr);
        gui_clear_status();
        gui_set_status(gettext("Sleeping ..."));
        VERIFY(0 == pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL));
        DEBUG_MTTHR(DBGvars("mt_semaphore_up(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
        mt_semaphore_up(&cfg.nrunning_sem);
        /*
         * UN-critical section; also signal that this thread isn't really active now.
         *
         * The effect of this is that the main thread can monitor if anybody is still
         * working on some URL; if all threads are waiting in the next loop for another
         * URL to arrive, the main thread will discover this after a while and will then
         * be able to signal all threads to stop as everything is done.
         */

        DEBUG_MTTHR(DBGvars("mt_semaphore_timed_wait(cfg.urlstack_sem: v = %ld BEFORE)\n"), cfg.urlstack_sem.v);
        while (!cfg.stop && !cfg.rbreak && mt_semaphore_timed_wait(&cfg.urlstack_sem, SUBSECOND_SEMA_POOL_TIMEOUT) < 0)
        {
      DEBUG_MTTHR(DBGvars("recurse: [%d/%ld] wait for cfg.urlstack_sem: v = %ld\n"),
                  thrnr,
                  cfg.nthr,
                  cfg.urlstack_sem.v);
        }

        DEBUG_MTTHR(DBGvars("mt_semaphore_decrement(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
        mt_semaphore_decrement(&cfg.nrunning_sem);
    }
    DEBUG_MTTHR(DBGvars("mt_semaphore_up(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
    mt_semaphore_up(&cfg.nrunning_sem);

#ifdef I_FACE
    pthread_cleanup_pop(TRUE);
#endif
    DEBUG_MTTHR(DBGvars("thread %d exiting\n"), thrnr);
    gui_set_status(gettext("Exiting ..."));

    DEBUG_PROCE(NULL);

    pthread_exit(NULL);
    return NULL;
}

void recurse(int dumb)
{
    pthread_attr_t thrdattr;
    int i;
    int num = cfg.nthr;

    DEBUG_PROCS(NULL);

    {
#if defined(HAVE_SIGEMPTYSET) && defined(HAVE_SIGADDSET) && defined(HAVE_PTHREAD_SIGMASK)
        sigset_t smask;

        sigemptyset(&smask);
        sigaddset(&smask, SIGINT);
#ifdef SIGQUIT
        sigaddset(&smask, SIGQUIT);
#endif
#ifdef SIGXFSZ
        sigaddset(&smask, SIGXFSZ);
#endif
        pthread_sigmask(SIG_UNBLOCK, &smask, NULL);
#endif

        signal(SIGINT, pavuk_sigintthr);
#ifdef SIGQUIT
        signal(SIGQUIT, pavuk_sigquitthr);
#endif
#ifdef SIGXFSZ
        signal(SIGXFSZ, pavuk_sigfilesizethr);
#endif
    }

    VERIFY(0 == pthread_attr_init(&thrdattr));
    VERIFY(0 == pthread_attr_setscope(&thrdattr, PTHREAD_SCOPE_SYSTEM));
    VERIFY(0 == pthread_attr_setstacksize(&thrdattr, MT_STACK_SIZE));

    DEBUG_MTTHR(DBGvars("mt_semaphore_init(nrunning/urlstack)\n"));
    mt_semaphore_destroy(&cfg.nrunning_sem);
    mt_semaphore_destroy(&cfg.urlstack_sem);
    DEBUG_MTTHR(DBGvars("mt_semaphore_init(cfg.nrunning_sem = 1)\n"));
    mt_semaphore_initv(&cfg.nrunning_sem, 1);
    DEBUG_MTTHR(DBGvars("mt_semaphore_init(cfg.urlstack_sem = 0)\n"));
    mt_semaphore_initv(&cfg.urlstack_sem, 0);
    /* signal each URL available in the queue */
    {
        dllist *dptr;

        LOCK_CFG_URLSTACK;
        DEBUG_MTTHR(DBGvars("multiple mt_semaphore_up(cfg.urlstack_sem: v = %ld BEFORE)\n"), cfg.urlstack_sem.v);
        for (dptr = cfg.urlstack; dptr; dptr = dptr->next)
        {
            mt_semaphore_up(&cfg.urlstack_sem);
        }
        DEBUG_MTTHR(DBGvars("multiple mt_semaphore_up(cfg.urlstack_sem: v = %ld AFTERWARDS)\n"), cfg.urlstack_sem.v);
        UNLOCK_CFG_URLSTACK;
    }

    if (num <= 0)
    {
        num = 1;
    }

    cfg.allthreadsnr = 0;
    cfg.allthreads = _malloc(num * sizeof(cfg.allthreads[0]));

    for (i = 0; (i < num) && !cfg.stop && !cfg.rbreak; i++)
    {
        if (!pthread_create(&cfg.allthreads[cfg.allthreadsnr], &thrdattr, &mt_recurse_thrd, cast_int2voidptr(cfg.allthreadsnr)))
        {
            cfg.allthreadsnr++;
#if defined I_FACE
            gui_mt_thread_start(cfg.allthreadsnr);
#endif
            DEBUG_MTTHR(DBGvars("mt_semaphore_decrement(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
            mt_semaphore_decrement(&cfg.nrunning_sem);
        }
        else
        {
            xvperror(DBGvars(NULL), gettext("Create downloading thread %d\n"), i);
        }
    }

    DEBUG_MTTHR(DBGvars("mt_semaphore_timed_wait(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
    while (!cfg.stop && !cfg.rbreak && mt_semaphore_timed_wait(&cfg.nrunning_sem, SUBSECOND_SEMA_POOL_TIMEOUT) < 0)
    {
      DEBUG_MTTHR(DBGvars("recurse: [-/%d] wait for cfg.nrunning_sem: v = %ld\n"),
                  num,
                  cfg.nrunning_sem.v);
    }

    DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "recurse (after waiting for cfg.running_sem) - ", cfg.stop, cfg.rbreak);

    /*
     * signal all processing threads to stop as each one of them is only waiting
     * for another URL to arrive in the URL stack/queue right now - which will never
     * happen as all URLs have been processed.
     */
    cfg.stop = TRUE;

    tl_msleep(SUBSECOND_SEMA_POOL_TIMEOUT, TRUE);

    for (i = 0; i < cfg.allthreadsnr; i++)
    {
#if 0
        pthread_cancel(cfg.allthreads[i]);
        pthread_kill(cfg.allthreads[i], SIGQUIT);
#endif
        pthread_join(cfg.allthreads[i], NULL);
    }

#if 0
    DEBUG_MTTHR(DBGvars("mt_semaphore_destroy(cfg.nrunning_sem)\n"));
    mt_semaphore_destroy(&cfg.nrunning_sem);
#endif
    _free(cfg.allthreads);
    cfg.allthreadsnr = 0;
#if defined I_FACE
    gui_mt_thread_end(0);
#endif

    DEBUG_PROCE(NULL);
}
#endif

void dump_ftp_list(dllist *urllst)
{
    dllist *ptr = urllst;

    DEBUG_PROCS(NULL);

    while (ptr)
    {
        url *urlp = (url *)ptr->data;
        void *dupl;

        dupl = dllist_find2(ptr->next, (dllist_t)urlp, dllist_url_compare);

    if (!dupl && !(urlp->status & URL_INLINE_OBJ) &&
        (urlp->type == URLT_FTP || urlp->type == URLT_FTPS))
        {
            const char *p, *pp;

            p = url_get_path(urlp);
            pp = strrchr(p, '/');
            if (pp)
            {
                pp++;
                if (!*pp)
                {
                    pp -= 2;
                    while (pp > p && *pp != '/')
                    {
                        pp--;
                    }
                    pp++;
                }
                if (urlp->extension)
                {
                    ftp_url_extension *fe = urlp->extension;

                    if (fe->type == FTP_TYPE_F)
                    {
                        xprintf(1, gettext("\t%s    (%ld bytes)\n"), pp, (long)fe->size);
                    }
                    else if (fe->type == FTP_TYPE_L)
                    {
                        xprintf(1, "\t%s    -> %s\n", pp, fe->slink);
                    }
                    else if (fe->type == FTP_TYPE_D)
                    {
                        xprintf(1, "\t%s/\n", pp);
                    }
                }
                else
                {
                    xprintf(1, "\t%s\n", pp);
                }
            }
        }
        free_deep_url(urlp);
        _free(urlp);
        ptr = dllist_remove_entry(ptr, ptr);
    }

    DEBUG_PROCE(NULL);
}






void dump_url(url *urlp)
{
    DEBUG_PROCS(NULL);

    if (bufio_is_open(cfg.dump_urlfd) || bufio_is_open(cfg.dump_fd))
    {
        char *urlstr = NULL;
        char *urldbgstr = NULL;

        cvt_url2diag_str(urlp, &urldbgstr, &urlstr);

        if (bufio_is_open(cfg.dump_urlfd))
        {
            bufio_puts_m(cfg.dump_urlfd, urlstr, "\n", NULL);
        }
        if (bufio_is_open(cfg.dump_fd))
        {
            bufio_puts_m(cfg.dump_fd, urldbgstr, "\n", NULL);
        }
        _free(urlstr);
        _free(urldbgstr);
    }

    DEBUG_PROCE(NULL);
}


void dump_urls_list(dllist *urls, doc *docu)
{
    dllist *ptr;
    void *dupl;

    DEBUG_PROCS(NULL);

    if (bufio_is_open(cfg.dump_urlfd) || bufio_is_open(cfg.dump_fd))
    {
        if (bufio_is_open(cfg.dump_urlfd))
        {
            bufio_puts(cfg.dump_urlfd, "\n---------------------- URL DUMP ----------------------\n");
        }
        if (bufio_is_open(cfg.dump_fd))
        {
            bufio_puts(cfg.dump_fd, "\n---------------------- URL DUMP ----------------------\n");
        }

        dump_url(docu->doc_url);
        if (bufio_is_open(cfg.dump_urlfd))
        {
            bufio_puts(cfg.dump_urlfd, "  -->\n");
        }
        if (bufio_is_open(cfg.dump_fd))
        {
            bufio_puts(cfg.dump_fd, "  -->\n");
        }

        for (ptr = urls; ptr; ptr = ptr->next)
        {
            dupl = dllist_find2(ptr->next, (dllist_t)ptr->data, dllist_url_compare);
            if (!dupl)
            {
                dump_url((url *)ptr->data);
            }
        }
    }

    DEBUG_PROCE(NULL);
}

void dump_cfg_formdata(const char *urlstr, dllist *formdata)
{
    dllist *uptr;
    dllist *fptr;
    cookie_entry *cptr;

    DEBUG_PROCS(NULL);

    if (bufio_is_open(cfg.dump_fd))
    {
        bufio *fd;

#ifdef HAVE_MT
        ASSERT(cfg.dump_fd->mutex);
#endif
        VERIFY(!bufio_lock(cfg.dump_fd, FALSE, &fd));
        for (uptr = formdata; uptr; uptr = uptr->next)
        {
            url_info *ui = (url_info *)uptr->data;
            if (ui)
            {
                char *method_str = "unknown";
                char *enc_str = "unknown";
                if (ui->method == FORM_M_GET)
                {
                    method_str = "GET";
                }
                else if (ui->method == FORM_M_POST)
                {
                    method_str = "POST";
                }
                if (ui->encoding == FORM_E_URLENCODED)
                {
                    enc_str = "application/x-www-form-urlencoded";
                }
                else if (ui->encoding == FORM_E_MULTIPART)
                {
                    enc_str = "multipart/form-data";
                }
                bufio_printf(fd, "--formdata: URL: '%s'\n", urlstr);
                bufio_printf(fd, "--          urlstr: '%s'\n", (ui->urlstr ? ui->urlstr : "(null)"));
                bufio_printf(fd, "--          type: '%d'\n", ui->type);
                bufio_printf(fd, "--          method: '%s'\n", method_str);
                bufio_printf(fd, "--          encoding: '%s'\n", enc_str);
                bufio_printf(fd, "--          fields:\n");
                fptr = ui->fields;
                while (fptr)
                {
                    char *p3[3];
                    form_field *ff = (form_field *)fptr->data;

                    switch (ff->type)
                    {
                    case FORM_T_FILE:
                        p3[0] = "FILE";
                        break;

                    case FORM_T_TEXTAREA:
                        p3[0] = "LONG TEXT";
                        break;

                    default:
                        p3[0] = "TEXT";
                    }

                    p3[1] = form_encode_urlencoded_str(ff->name);
                    p3[2] = form_encode_urlencoded_str(ff->value);

          bufio_printf(fd,
                       "--          - field: type: %s, name: '%s', value: '%s'\n",
                       p3[0],
                       p3[1],
                       p3[2]);

                    _free(p3[1]);
                    _free(p3[2]);
                    fptr = fptr->next;
                }
                cptr = ui->cookies;
                ASSERT(cptr ? !cptr->prev : TRUE);
                while (cptr)
                {
                    char *cmsg;

                    cmsg = cvt_cookie_to_setstr(cptr);
                    bufio_printf(fd, "--          - cookie: %s\n", cmsg);
                    _free(cmsg);

                    cptr = cptr->next;
                }
        bufio_printf(fd, "--          localname: '%s'\n",
                     (ui->localname ? ui->localname : "(null)"));
            }
        }
        VERIFY(!bufio_release(cfg.dump_fd, &fd));
    }

    DEBUG_PROCE(NULL);
}

void dump_formlist(const char *urlstr, dllist *formlist)
{
    dllist *fptr;
    int nform;

    DEBUG_PROCS(NULL);

    if (bufio_is_open(cfg.dump_fd))
    {
        bufio *fd;

#ifdef HAVE_MT
        ASSERT(cfg.dump_fd->mutex);
#endif
        VERIFY(!bufio_lock(cfg.dump_fd, FALSE, &fd));
        for (fptr = formlist, nform = 0; fptr; fptr = fptr->next, nform++)
        {
            url *urlp;
            char *typ_str = "???";

            bufio_printf(fd, "-- formlist: RAW URL: '%s'\n", (char *)fptr->data);
            DEBUG_TRACE("dump_formlist:url_parse:enter\n");
            urlp = url_parse((char *)fptr->data);
            DEBUG_TRACE("dump_formlist:url_parse:leave\n");
            /* ASSERT(urlp->type != URLT_FROMPARENT); */

            if (urlp)
            {
                switch (urlp->type)
                {
                case URLT_HTTP:
                    typ_str = "HTTP";
                    break;

                case URLT_HTTPS:
                    typ_str = "HTTPS";
                    break;

                case URLT_FTP:
                    typ_str = "FTP";
                    break;

                case URLT_FTPS:
                    typ_str = "FTPS";
                    break;

                case URLT_FILE:
                    typ_str = "FILE";
                    break;

                case URLT_GOPHER:
                    typ_str = "GOPHER";
                    break;

                case URLT_FROMPARENT:
                    typ_str = "FROMPARENT";
                    break;

                default:
                    typ_str = "UNKNOWN";
                    break;
                }
            }
            bufio_printf(fd, "--           TYPE: %s\n", typ_str);
            if (urlp)
            {
                LOCK_URL(urlp);
                if (urlp->parent_url)
                {
                    bufio_printf(fd, "--           has parent_utl\n");
                }
                if (urlp->moved_to)
                {
                    bufio_printf(fd, "--           has moved to: '%s'\n", urlp->moved_to->local_name);
                }
                bufio_printf(fd, "--           level: %d\n", urlp->level);
                bufio_printf(fd, "--           ref count: %d\n", urlp->ref_cnt);
                UNLOCK_URL(urlp);
                bufio_printf(fd, "--           status bits: 0x%08X\n", urlp->status);
                bufio_printf(fd, "--           local name: '%s'\n", urlp->local_name);
                bufio_printf(fd, "--           extension: %p\n", urlp->extension);
            }
            free_deep_url(urlp);
            _free(urlp);
        }
        VERIFY(!bufio_release(cfg.dump_fd, &fd));
    }

    DEBUG_PROCE(NULL);
}

void get_urls_to_resume(const char *dirname)
{
    DIR *dir;
    struct dirent *dent;
    char next_dir[PATH_MAX];
    struct stat estat;
    url *purl;

    DEBUG_PROCS(NULL);

    if (!(dir = tl_opendir(dirname)))
    {
        xperror1(gettext("Cannot open/read directory"), mk_native(dirname));

        DEBUG_PROCE(NULL);
        return;
    }

    gui_set_msg(gettext("Searching for files to resume"), 0);

    while ((dent = readdir(dir)) && !cfg.stop && !cfg.rbreak)
    {
        gui_loop_serve();

        snprintf(next_dir, sizeof(next_dir), "%s%s%s", dirname, (tl_is_dirname(dirname) ? "" : "/"), dent->d_name);
        next_dir[sizeof(next_dir) - 1] = 0;

        if (!strcmp(dent->d_name, "."))
            continue;
        if (!strcmp(dent->d_name, ".."))
            continue;

        if (tl_lstat(next_dir, &estat))
        {
            xperror(mk_native(next_dir));
            continue;
        }

        if (S_ISDIR(estat.st_mode))
        {
            if (!strcmp(dent->d_name, PAVUK_INFO_DIRNAME) && cfg.enable_info)
            {
                continue;
            }

            get_urls_to_resume(next_dir);
        }
        else if (!strncmp(PAVUK_FNAME_PREFIX, dent->d_name, strlen(PAVUK_FNAME_PREFIX)))
        {
            snprintf(next_dir, sizeof(next_dir), "%s%s%s", dirname, (tl_is_dirname(dirname) ? "" : "/"),
                     dent->d_name + strlen(PAVUK_FNAME_PREFIX));
            next_dir[sizeof(next_dir) - 1] = 0;
            if ((purl = filename_to_url(next_dir)))
            {
                if (cfg.mode != MODE_MIRROR)
                {
                    xprintf(1, gettext("Adding '%s' to resume list\n"), next_dir);
                }
                purl->status |= URL_ISSTARTING;
                url_set_filename(purl, tl_strdup(next_dir));
                append_url_to_list(purl);
            }
        }
    }

    closedir(dir);

    DEBUG_PROCE(NULL);
}

void get_urls_to_synchronize(const char *dirname, dllist **list)
{
    DIR *dir;
    struct dirent *dent;
    char next_dir[PATH_MAX];
    struct stat estat;
    url *purl;

    DEBUG_PROCS(NULL);

    if (!(dir = tl_opendir(dirname)))
    {
        xperror1(gettext("Cannot open/read directory"), mk_native(dirname));
        return;
    }

    gui_set_msg(gettext("Searching for documents to synchronize"), 0);

    while ((dent = readdir(dir)) && !cfg.stop && !cfg.rbreak)
    {
        gui_loop_serve();

        snprintf(next_dir, sizeof(next_dir), "%s%s%s", dirname, (tl_is_dirname(dirname) ? "" : "/"), dent->d_name);
        next_dir[sizeof(next_dir) - 1] = 0;

        if (!strcmp(dent->d_name, "."))
            continue;
        if (!strcmp(dent->d_name, ".."))
            continue;

        if (tl_lstat(next_dir, &estat))
        {
            xperror(mk_native(next_dir));
            continue;
        }

        if (S_ISDIR(estat.st_mode))
        {
            if (!strcmp(dent->d_name, PAVUK_INFO_DIRNAME) && cfg.enable_info)
            {
                continue;
            }

            strcat(next_dir, "/");

            if ((purl = filename_to_url(next_dir)) && (purl->type == URLT_FTP || purl->type == URLT_FTPS) && !cfg.store_index)
            {
                char *ustr;

                purl->status |= URL_ISSTARTING;
                purl->extension = ftp_url_ext_new(FTP_TYPE_D, -1, -1, NULL, 0);
                url_set_filename(purl, tl_str_concat(DBGvars(NULL), next_dir, priv_cfg.index_name, NULL));
                *list = dllist_prepend(*list, (dllist_t)purl);

                ustr = url_to_urlstr(purl, FALSE);
                if (cfg.mode != MODE_MIRROR)
                {
                    xprintf(1, gettext("Adding directory '%s' to sync list as URL %s\n"), next_dir, ustr);
                }
                _free(ustr);
            }
            else if (purl)
            {
                free_deep_url(purl);
                _free(purl);
            }

            next_dir[strlen(next_dir) - 1] = '\0';

            get_urls_to_synchronize(next_dir, list);
        }
        else if (cfg.enable_info && !strcmp(dent->d_name, PAVUK_LOCK_FILENAME))
        {
            /* do nothing */
            continue;
        }
        else if (!strncmp(PAVUK_FNAME_PREFIX, dent->d_name, strlen(PAVUK_FNAME_PREFIX)))
        {
            snprintf(next_dir, sizeof(next_dir), "%s%s%s", dirname, (tl_is_dirname(dirname) ? "" : "/"),
                     dent->d_name + strlen(PAVUK_FNAME_PREFIX));
            next_dir[sizeof(next_dir) - 1] = 0;
            if ((purl = filename_to_url(next_dir)))
            {
                char *ustr;

                ustr = url_to_urlstr(purl, FALSE);
                if (cfg.mode != MODE_MIRROR)
                {
                    xprintf(1, gettext("Adding file '%s' to sync list as URL %s\n"), next_dir, ustr);
                }
                _free(ustr);
                if (purl->type == URLT_FTP || purl->type == URLT_FTPS)
                {
                    int tp;

                    if (purl->p.ftp.dir)
                    {
                        tp = FTP_TYPE_D;
                    }
                    else
                    {
                        tp = FTP_TYPE_F;
                    }
                    purl->extension = ftp_url_ext_new(tp, -1, -1, NULL, 0);
                }
                purl->status |= URL_ISSTARTING;

                url_set_filename(purl, tl_strdup(next_dir));
                *list = dllist_prepend(*list, (dllist_t)purl);
            }
        }
        else
        {
            if ((purl = filename_to_url(next_dir)))
            {
                char *ustr;

                ustr = url_to_urlstr(purl, FALSE);
                if (cfg.mode != MODE_MIRROR)
                {
                    xprintf(1, gettext("Adding file '%s' to sync list as URL %s\n"), next_dir, ustr);
                }
                _free(ustr);

                if (purl->type == URLT_FTP)
                {
                    int tp;

                    if (purl->p.ftp.dir)
                    {
                        tp = FTP_TYPE_D;
                    }
#ifdef S_ISLNK
                    else if (S_ISLNK(estat.st_mode))
                    {
                        tp = FTP_TYPE_L;
                    }
#endif
                    else
                    {
                        tp = FTP_TYPE_F;
                    }
                    purl->extension = ftp_url_ext_new(tp, -1, -1, NULL, 0);
                }

                purl->status |= URL_ISSTARTING;
                url_set_filename(purl, tl_strdup(next_dir));
                *list = dllist_prepend(*list, (dllist_t)purl);
            }
        }
    }

    closedir(dir);

    DEBUG_PROCE(NULL);
}

